import numpy as np
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from ..utils import comprehend_axes
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item


class FourierLibrary(BaseFeatureLibrary):
    """
    Generate a library with trigonometric functions.

    Parameters
    ----------
    n_frequencies : int, optional (default 1)
        Number of frequencies to include in the library. The library will
        include functions :math:`\\sin(x), \\sin(2x), \\dots
        \\sin(n_{frequencies}x)` for each input feature :math:`x`
        (depending on which of sine and/or cosine features are included).

    include_sin : boolean, optional (default True)
        If True, include sine terms in the library.

    include_cos : boolean, optional (default True)
        If True, include cosine terms in the library.

    Attributes
    ----------
    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is ``2 * n_input_features_ * n_frequencies`` if both sines and cosines
        are included. Otherwise it is ``n_input_features * n_frequencies``.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import FourierLibrary
    >>> x = np.array([[0.],[1.],[2.]])
    >>> lib = FourierLibrary(n_frequencies=2).fit(x)
    >>> lib.transform(x)
    array([[ 0.        ,  1.        ,  0.        ,  1.        ],
           [ 0.84147098,  0.54030231,  0.90929743, -0.41614684],
           [ 0.90929743, -0.41614684, -0.7568025 , -0.65364362]])
    >>> lib.get_feature_names()
    ['sin(1 x0)', 'cos(1 x0)', 'sin(2 x0)', 'cos(2 x0)']
    """

    def __init__(
        self,
        n_frequencies=1,
        include_sin=True,
        include_cos=True,
    ):
        if not (include_sin or include_cos):
            raise ValueError("include_sin and include_cos cannot both be False")
        if n_frequencies < 1 or not isinstance(n_frequencies, int):
            raise ValueError("n_frequencies must be a positive integer")
        self.n_frequencies = n_frequencies
        self.include_sin = include_sin
        self.include_cos = include_cos

    def get_feature_names(self, input_features=None):
        """
        Return feature names for output features

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)
        n_input_features = self.n_features_in_
        if input_features is None:
            input_features = ["x%d" % i for i in range(n_input_features)]
        feature_names = []
        for i in range(self.n_frequencies):
            for feature in input_features:
                if self.include_sin:
                    feature_names.append("sin(" + str(i + 1) + " " + feature + ")")
                if self.include_cos:
                    feature_names.append("cos(" + str(i + 1) + " " + feature + ")")
        return feature_names

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data.

        Returns
        -------
        self : instance
        """
        n_features = x_full[0].shape[x_full[0].ax_coord]
        self.n_features_in_ = n_features
        if self.include_sin and self.include_cos:
            self.n_output_features_ = n_features * self.n_frequencies * 2
        else:
            self.n_output_features_ = n_features * self.n_frequencies
        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data to Fourier features

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape (n_samples, n_output_features)
            The matrix of features, where n_output_features is the number of Fourier
            features generated from the inputs.
        """
        check_is_fitted(self)

        xp_full = []
        for x in x_full:
            # n_samples = x.shape[x.ax_sample]
            n_features = x.shape[x.ax_coord]
            shape = np.array(x.shape)

            n_input_features = self.n_features_in_
            if n_features != n_input_features:
                raise ValueError("x shape does not match training shape")

            shape[-1] = self.n_output_features_
            xp = np.empty(shape, dtype=x.dtype)
            idx = 0
            for i in range(self.n_frequencies):
                for j in range(n_input_features):
                    if self.include_sin:
                        xp[..., idx] = np.sin((i + 1) * x[..., j])
                        idx += 1
                    if self.include_cos:
                        xp[..., idx] = np.cos((i + 1) * x[..., j])
                        idx += 1
            xp = AxesArray(xp, comprehend_axes(xp))
            xp_full.append(xp)
        return xp_full



"""Unit tests for SINDy with control."""
import numpy as np
import pytest
from scipy.interpolate import interp1d
from sklearn.exceptions import ConvergenceWarning
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lasso
from sklearn.utils.validation import check_is_fitted

from pysindy import DiscreteSINDy
from pysindy import SINDy
from pysindy.optimizers import SR3
from pysindy.optimizers import STLSQ


def test_get_feature_names_len(data_lorenz_c_1d):
    x, t, u, _ = data_lorenz_c_1d
    model = SINDy()

    model.fit(x, u=u, t=t)

    # Assumes default library is polynomial features of degree 2
    assert len(model.get_feature_names()) == 21


def test_not_fitted(data_lorenz_c_1d):
    x, t, u, u_fun = data_lorenz_c_1d
    model = SINDy()

    with pytest.raises(NotFittedError):
        model.predict(x, u=u)
    with pytest.raises(NotFittedError):
        model.simulate(x[0], t=t, u=u_fun, integrator_kws={"rtol": 0.1})


def test_improper_shape_input(data_1d):
    x, t = data_1d
    u = np.ones_like(x)

    # Ensure model successfully handles different data shapes
    model = SINDy()
    model.fit(x.flatten(), u=u, t=t)
    check_is_fitted(model)

    model = SINDy()
    model.fit(x.flatten(), u=u, t=t, x_dot=x.flatten())
    check_is_fitted(model)

    model = SINDy()
    model.fit(x, u=u, t=t, x_dot=x.flatten())
    check_is_fitted(model)

    model = SINDy()
    model.fit(x.flatten(), u=u.flatten(), t=t)
    check_is_fitted(model)

    model = SINDy()
    model.fit(x.flatten(), u=u.flatten(), t=t, x_dot=x.flatten())
    check_is_fitted(model)

    model = SINDy()
    model.fit(x, u=u.flatten(), t=t, x_dot=x.flatten())
    check_is_fitted(model)

    # Should fail if x and u have incompatible numbers of rows
    with pytest.raises(ValueError):
        model.fit(x[:-1, :], u=u, t=t[:-1])


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_lorenz_c_1d"), pytest.lazy_fixture("data_lorenz_c_2d")],
)
def test_mixed_inputs(data):
    x, t, u, _ = data

    # Scalar t
    model = SINDy()
    model.fit(x, u=u, t=2)
    check_is_fitted(model)

    model = SINDy()
    model.fit(x, u=u, t=t, x_dot=x)
    check_is_fitted(model)


def test_bad_control_input(data_lorenz_c_1d):
    x, t, u, _ = data_lorenz_c_1d
    model = SINDy()

    with pytest.raises(TypeError):
        model.fit(x, u=set(u), t=t)


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_lorenz_c_1d"), pytest.lazy_fixture("data_lorenz_c_2d")],
)
def test_bad_t(data):
    x, t, u, _ = data
    model = SINDy()

    # Wrong type
    with pytest.raises(ValueError):
        model.fit(x, u=u, t="1")

    # Invalid value of t
    with pytest.raises(ValueError):
        model.fit(x, u=u, t=-1)

    # t is a list
    with pytest.raises(ValueError):
        model.fit(x, u=u, t=list(t))

    # Wrong number of time points
    with pytest.raises(ValueError):
        model.fit(x, u=u, t=t[:-1])

    t_new = np.copy(t)
    # Two points in t out of order
    t_new[2], t_new[4] = t_new[4], t_new[2]
    with pytest.raises(ValueError):
        model.fit(x, u=u, t=t_new)
    t_new[2], t_new[4] = t_new[4], t_new[2]

    # Two matching times in t
    t_new[3] = t_new[5]
    with pytest.raises(ValueError):
        model.fit(x, u=u, t=t_new)


@pytest.mark.parametrize(
    "data, optimizer",
    [
        (pytest.lazy_fixture("data_lorenz_c_1d"), STLSQ()),
        (pytest.lazy_fixture("data_lorenz_c_2d"), STLSQ()),
        (pytest.lazy_fixture("data_lorenz_c_1d"), SR3()),
        (pytest.lazy_fixture("data_lorenz_c_2d"), SR3()),
        (pytest.lazy_fixture("data_lorenz_c_1d"), Lasso(fit_intercept=False)),
        (pytest.lazy_fixture("data_lorenz_c_2d"), Lasso(fit_intercept=False)),
        (pytest.lazy_fixture("data_lorenz_c_1d"), ElasticNet(fit_intercept=False)),
        (pytest.lazy_fixture("data_lorenz_c_2d"), ElasticNet(fit_intercept=False)),
    ],
)
def test_predict(data, optimizer):
    x, t, u, _ = data
    model = SINDy(optimizer=optimizer)
    model.fit(x, u=u, t=t)
    x_dot = model.predict(x, u=u)

    assert x.shape == x_dot.shape


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_lorenz_c_1d"), pytest.lazy_fixture("data_lorenz_c_2d")],
)
def test_simulate(data):
    x, t, u, u_fun = data
    model = SINDy()
    model.fit(x, u=u, t=t)
    x1 = model.simulate(x[0], t=t, u=u_fun, integrator_kws={"rtol": 0.1})

    assert len(x1) == len(t)


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_lorenz_c_1d"), pytest.lazy_fixture("data_lorenz_c_2d")],
)
def test_simulate_with_interp(data):
    x, t, u, _ = data
    model = SINDy()
    model.fit(x, u=u, t=t)

    u_fun = interp1d(t, u, axis=0)
    x1 = model.simulate(x[0], t=t, u=u_fun, integrator_kws={"rtol": 0.1})

    assert len(x1) == len(t)


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_lorenz_c_1d"), pytest.lazy_fixture("data_lorenz_c_2d")],
)
def test_simulate_with_vector_control_input(data):
    x, t, u, _ = data
    model = SINDy()
    model.fit(x, u=u, t=t)

    x1 = model.simulate(x[0], t=t, u=u, integrator_kws={"rtol": 0.1})

    assert len(x1) == len(t) - 1


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_lorenz_c_1d"), pytest.lazy_fixture("data_lorenz_c_2d")],
)
def test_score(data):
    x, t, u, _ = data
    model = SINDy()
    model.fit(x, u=u, t=t)

    assert model.score(x, u=u, t=t) <= 1

    assert model.score(x, u=u, t=t, x_dot=x) <= 1


def test_fit_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    u = [np.ones((xi.shape[0], 2)) for xi in x]

    model = SINDy()

    model.fit(x, u=u, t=t)
    assert model.score(x, u=u, t=t) > 0.8

    model = SINDy()
    model.fit(x, u=u, t=t, x_dot=x)
    check_is_fitted(model)


def test_predict_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    u = [np.ones((xi.shape[0], 2)) for xi in x]

    model = SINDy()
    model.fit(x, u=u, t=t)

    p = model.predict(x, u=u)
    assert len(p) == len(x)


def test_score_multiple_trajectories(data_multiple_trajectories):
    x, t = data_multiple_trajectories
    u = [np.ones((xi.shape[0], 2)) for xi in x]

    model = SINDy()
    model.fit(x, u=u, t=t)
    check_is_fitted(model)

    s = model.score(x, u=u, t=t)
    assert s <= 1

    s = model.score(x, u=u, t=t, x_dot=x)
    assert s <= 1


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_discrete_time_c"),
        pytest.lazy_fixture("data_discrete_time_c_multivariable"),
    ],
)
def test_fit_discrete_time(data):
    x, u = data

    model = DiscreteSINDy()
    model.fit(x, u=u, t=1)
    check_is_fitted(model)

    model = DiscreteSINDy()
    model.fit(x[:-1], u=u[:-1], x_next=x[1:], t=1)
    check_is_fitted(model)


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_discrete_time_c"),
        pytest.lazy_fixture("data_discrete_time_c_multivariable"),
    ],
)
def test_simulate_discrete_time(data):
    x, u = data
    model = DiscreteSINDy()
    model.fit(x, u=u, t=1)
    n_steps = x.shape[0]
    x1 = model.simulate(x[0], t=n_steps, u=u)

    assert len(x1) == n_steps

    # TODO: implement test using the stop_condition option


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_discrete_time_c"),
        pytest.lazy_fixture("data_discrete_time_c_multivariable"),
    ],
)
def test_predict_discrete_time(data):
    x, u = data
    model = DiscreteSINDy()
    print(x, u)
    model.fit(x, u=u, t=1)
    assert len(model.predict(x, u=u)) == len(x)


@pytest.mark.parametrize(
    "data",
    [
        pytest.lazy_fixture("data_discrete_time_c"),
        pytest.lazy_fixture("data_discrete_time_c_multivariable"),
    ],
)
def test_score_discrete_time(data):
    x, u = data
    model = DiscreteSINDy()
    model.fit(x, u=u, t=1)
    assert model.score(x, u=u, t=1) > 0.75
    assert model.score(x, u=u, x_next=x, t=1) < 1


def test_fit_discrete_time_multiple_trajectories(
    data_discrete_time_multiple_trajectories_c,
):
    x, u = data_discrete_time_multiple_trajectories_c
    model = DiscreteSINDy()
    model.fit(x, u=u, t=1)
    check_is_fitted(model)

    model = DiscreteSINDy()
    model.fit(x, u=u, x_next=x, t=1)
    check_is_fitted(model)


def test_predict_discrete_time_multiple_trajectories(
    data_discrete_time_multiple_trajectories_c,
):
    x, u = data_discrete_time_multiple_trajectories_c
    model = DiscreteSINDy()
    model.fit(x, u=u, t=1)

    y = model.predict(x, u=u)
    assert len(y) == len(x)


def test_score_discrete_time_multiple_trajectories(
    data_discrete_time_multiple_trajectories_c,
):
    x, u = data_discrete_time_multiple_trajectories_c
    model = DiscreteSINDy()
    model.fit(x, u=u, t=1)

    s = model.score(x, u=u, t=1)
    assert s > 0.75

    # x is not its own derivative, so we expect bad performance here
    s = model.score(x, u=u, x_next=x, t=1)
    assert s < 1


def test_simulate_errors(data_lorenz_c_1d):
    x, t, u, u_fun = data_lorenz_c_1d
    model = SINDy()
    model.fit(x, u=u, t=t)

    with pytest.raises(ValueError):
        model.simulate(x[0], t=1, u=u)

    model = DiscreteSINDy()
    with pytest.raises(ValueError):
        model.simulate(x[0], t=[1, 2], u=u)


@pytest.mark.parametrize(
    "params, warning",
    [({"threshold": 100}, UserWarning), ({"max_iter": 1}, ConvergenceWarning)],
)
def test_fit_warn(data_lorenz_c_1d, params, warning):
    x, t, u, _ = data_lorenz_c_1d
    model = SINDy(optimizer=STLSQ(**params))

    with pytest.warns(warning):
        model.fit(x, u=u, t=t)


def test_u_omitted(data_lorenz_c_1d):
    x, t, u, _ = data_lorenz_c_1d
    model = SINDy()

    model.fit(x, u=u, t=t)

    with pytest.raises(TypeError):
        model.predict(x)

    with pytest.raises(TypeError):
        model.score(x)

    with pytest.raises(TypeError):
        model.simulate(x[0], t=t)


def test_extra_u_warn(data_lorenz_c_1d):
    x, t, u, _ = data_lorenz_c_1d
    model = SINDy()
    model.fit(x, t=t)

    with pytest.warns(UserWarning):
        model.predict(x, u=u)

    with pytest.warns(UserWarning):
        model.score(x, u=u, t=t)

    with pytest.warns(UserWarning):
        model.simulate(x[0], t=t, u=u, integrator_kws={"rtol": 0.1})


def test_extra_u_warn_discrete(data_discrete_time_c):
    x, u = data_discrete_time_c
    model = DiscreteSINDy()
    model.fit(x, t=1)

    with pytest.warns(UserWarning):
        model.predict(x, u=u)

    with pytest.warns(UserWarning):
        model.score(x, u=u, t=1)

    with pytest.warns(UserWarning):
        model.simulate(x[0], u=u, t=10)



from typing import cast
from typing import NewType
from typing import TypeVar

import numpy as np
from numpy.typing import NBitBase
from sklearn.linear_model import ridge_regression

from .base import BaseOptimizer

Rows = TypeVar("Rows", covariant=True, bound=int)
Cols = TypeVar("Cols", covariant=True, bound=int)
Float2D = np.ndarray[tuple[Rows, Cols], np.dtype[np.floating[NBitBase]]]
Features = NewType("Features", int)
Targets = NewType("Targets", int)
Samples = NewType("Samples", int)


class SSR(BaseOptimizer):
    """Stepwise sparse regression (SSR) greedy algorithm.

    Attempts to minimize the objective function
    :math:`\\|y - Xw\\|^2_2 + \\alpha \\|w\\|^2_2`
    by iteratively eliminating the smallest coefficient

    See the following reference for more details:

        Boninsegna, Lorenzo, Feliks Nüske, and Cecilia Clementi.
        "Sparse learning of stochastic dynamical equations."
        The Journal of chemical physics 148.24 (2018): 241723.

    Parameters
    ----------

    max_iter : int, optional (default 20)
        Maximum iterations of the optimization algorithm.

    normalize_columns : boolean, optional (default False)
        Normalize the columns of x (the SINDy library terms) before regression
        by dividing by the L2-norm. Note that the 'normalize' option in sklearn
        is deprecated in sklearn versions >= 1.0 and will be removed.

    copy_X : boolean, optional (default True)
        If True, X will be copied; else, it may be overwritten.

    kappa : float, optional (default None)
        If passed, compute the MSE errors with an extra L0 term with
        strength equal to kappa times the condition number of Theta.

    criteria : string, optional (default "coefficient_value")
        The criteria to use for truncating a coefficient each iteration.
        Must be "coefficient_value" or "model_residual".
        "coefficient_value": zero out the smallest coefficient).
        "model_residual": choose the N-1 term model with the smallest
        residual error.

    alpha : float, optional (default 0.05)
        Optional L2 (ridge) regularization on the weight vector.

    ridge_kw : dict, optional (default None)
        Optional keyword arguments to pass to the ridge regression.

    verbose : bool, optional (default False)
        If True, prints out the different error terms every iteration.

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s).

    history_ : list
        History of ``coef_``. ``history_[k]`` contains the values of
        ``coef_`` at iteration k of SSR

    err_history_ : list
        History of ``coef_``. ``history_[k]`` contains the MSE of each
        ``coef_`` at iteration k of SSR

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.integrate import odeint
    >>> from pysindy import SINDy
    >>> from pysindy.optimizers import SSR
    >>> lorenz = lambda z,t : [10 * (z[1] - z[0]),
    >>>                        z[0] * (28 - z[2]) - z[1],
    >>>                        z[0] * z[1] - 8 / 3 * z[2]]
    >>> t = np.arange(0, 2, .002)
    >>> x = odeint(lorenz, [-8, 8, 27], t)
    >>> opt = SSR(alpha=.5)
    >>> model = SINDy(optimizer=opt)
    >>> model.fit(x, t=t[1] - t[0])
    >>> model.print()
    x0' = -9.999 1 + 9.999 x0
    x1' = 27.984 1 + -0.996 x0 + -1.000 1 x1
    x2' = -2.666 x1 + 1.000 1 x0
    """

    def __init__(
        self,
        alpha=0.05,
        max_iter=20,
        ridge_kw=None,
        normalize_columns=False,
        copy_X=True,
        criteria="coefficient_value",
        kappa=None,
        verbose=False,
        unbias=True,
    ):
        super(SSR, self).__init__(
            max_iter=max_iter,
            copy_X=copy_X,
            normalize_columns=normalize_columns,
            unbias=unbias,
        )

        if alpha < 0:
            raise ValueError("alpha cannot be negative")

        if max_iter <= 0:
            raise ValueError("max iteration must be > 0")

        if criteria != "coefficient_value" and criteria != "model_residual":
            raise ValueError(
                "The only implemented criteria for sparsifying models "
                " are coefficient_value (zeroing out the smallest coefficient)"
                " or model_residual (choosing the N-1 term model with)"
                " the smallest residual error."
            )

        self.criteria = criteria
        self.alpha = alpha
        self.ridge_kw = ridge_kw
        self.kappa = kappa
        self.verbose = verbose

    def _coefficient_value(self, coef):
        """Eliminate the smallest element of the weight vector(s)"""
        c = coef
        inds_nonzero = np.ravel(np.nonzero(c))
        c_nonzero = c[inds_nonzero]
        smallest_ind = np.argmin(np.abs(c_nonzero))
        c[inds_nonzero[smallest_ind]] = 0.0
        return c, inds_nonzero[smallest_ind]

    def _model_residual(self, x, y, coef, inds):
        """Choose model with lowest residual error"""
        x_shape = np.shape(x)[-1]
        c = np.zeros((x_shape, x_shape - 1))
        err = np.zeros(x_shape)
        for i in range(x_shape):
            mask = np.ones(x_shape, dtype=bool)
            mask[i] = False
            c[i, :] = self._regress(x[:, mask], y)
            err[i] = np.sum((y - x[:, mask] @ c[i, :]) ** 2)
        min_err = np.argmin(err)
        # Figure out where this index is in the larger coef matrix
        total_ind = min_err
        q = -1
        for i in range(len(inds)):
            if q == min_err:
                break
            if not inds[i]:
                total_ind += 1
            else:
                q = q + 1
        cc = coef
        cc[total_ind] = 0.0
        return cc, total_ind

    def _regress(
        self, x: Float2D[Samples, Features], y: Float2D[Samples, Targets]
    ) -> Float2D[Targets, Features]:
        """Perform the ridge regression"""
        kw = self.ridge_kw or {}
        coef = ridge_regression(x, y, self.alpha, **kw)
        coef = np.atleast_2d(coef)  # type: ignore
        self.iters += 1
        return coef

    def _reduce(self, x, y):
        """Performs at most ``self.max_iter`` iterations of the
        SSR greedy algorithm.
        """
        # Until static typing grows, use cast
        x = cast(Float2D[Samples, Features], x)
        y = cast(Float2D[Samples, Targets], y)
        n_samples, n_features = x.shape
        n_targets = y.shape[1]
        cond_num = np.linalg.cond(x)
        if self.kappa is not None:
            l0_penalty = self.kappa * cond_num
        else:
            l0_penalty = 0

        coef = self._regress(x, y)
        inds = np.ones((n_targets, n_features), dtype=bool)

        # Print initial values for each term in the optimization
        if self.verbose:
            row = [
                "Iteration",
                "|y - Xw|^2",
                "a * |w|_2",
                "|w|_0",
                "b * |w|_0",
                "Total: |y-Xw|^2+a*|w|_2+b*|w|_0",
            ]
            print(
                "{: >10} ... {: >10} ... {: >10} ... {: >10}"
                " ... {: >10} ... {: >10}".format(*row)
            )

        self.history_ = [coef]
        self.err_history_ = [
            np.sum((y - x @ coef.T) ** 2) + l0_penalty * np.count_nonzero(coef)
        ]
        for k in range(self.max_iter):
            for i in range(n_targets):
                if self.criteria == "coefficient_value":
                    coef[i, :], ind = self._coefficient_value(coef[i, :])
                    inds[i, ind] = False
                    if np.any(inds[i, :]):
                        coef[i, inds[i, :]] = self._regress(x[:, inds[i, :]], y[:, i])
                else:
                    if np.sum(inds[i, :]) >= 2:
                        coef[i, :], ind = self._model_residual(
                            x[:, inds[i, :]], y[:, i], coef[i, :], inds[i, :]
                        )
                        inds[i, ind] = False
                        coef[i, inds[i, :]] = self._regress(x[:, inds[i, :]], y[:, i])

            self.history_.append(np.copy(coef))
            if self.verbose:
                R2 = np.sum((y - np.dot(x, coef.T)) ** 2)
                L2 = self.alpha * np.sum(coef**2)
                L0 = np.count_nonzero(coef)
                row = [k, R2, L2, L0, l0_penalty * L0, R2 + L2 + l0_penalty * L0]
                print(
                    "{0:10d} ... {1:10.4e} ... {2:10.4e} ... {3:10d}"
                    " ... {4:10.4e} ... {5:10.4e}".format(*row)
                )
            self.err_history_.append(
                np.sum((y - x @ coef.T) ** 2) + l0_penalty * np.count_nonzero(coef)
            )
            if np.all(np.sum(np.asarray(inds, dtype=int), axis=1) <= 1):
                # each equation has one last term
                break

        if self.kappa is not None:
            ind_best = np.argmin(self.err_history_)
        else:
            # err history is reverse of ordering in paper
            ind_best = (
                len(self.err_history_) - 1 - _ind_inflection(self.err_history_[::-1])
            )
        self.coef_ = np.asarray(self.history_)[ind_best, :, :]


def _ind_inflection(err_descending: list[float]) -> int:
    "Calculate the index of the inflection point in error"
    if len(err_descending) == 1:
        raise ValueError("Cannot find the inflection point of a single point")
    err_descending = np.array(err_descending)
    if np.any(err_descending < 0):
        raise ValueError("SSR inflection point method requires nonnegative losses")
    if np.any(err_descending == 0):
        return np.argmin(err_descending)
    err_ratio = err_descending[:-1] / err_descending[1:]
    return np.argmax(err_ratio) + 1



"""
Unit tests for optimizers.
"""
import pickle

import cvxpy as cp
import numpy as np
import pytest
from numpy.linalg import norm
from numpy.typing import NDArray
from scipy.integrate import solve_ivp
from sklearn.base import BaseEstimator
from sklearn.exceptions import ConvergenceWarning
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lasso
from sklearn.utils._param_validation import InvalidParameterError
from sklearn.utils.validation import check_is_fitted

from pysindy import FiniteDifference
from pysindy import PolynomialLibrary
from pysindy import SINDy
from pysindy.feature_library import CustomLibrary
from pysindy.feature_library import SINDyPILibrary
from pysindy.optimizers import BaseOptimizer
from pysindy.optimizers import ConstrainedSR3
from pysindy.optimizers import EnsembleOptimizer
from pysindy.optimizers import FROLS
from pysindy.optimizers import MIOSR
from pysindy.optimizers import SBR
from pysindy.optimizers import SINDyPI
from pysindy.optimizers import SR3
from pysindy.optimizers import SSR
from pysindy.optimizers import StableLinearSR3
from pysindy.optimizers import STLSQ
from pysindy.optimizers import TrappingSR3
from pysindy.optimizers import WrappedOptimizer
from pysindy.optimizers.base import _normalize_features
from pysindy.optimizers.ssr import _ind_inflection
from pysindy.optimizers.stlsq import _remove_and_decrement
from pysindy.utils import supports_multiple_targets
from pysindy.utils.odes import enzyme

# For reproducibility
np.random.seed(100)


class DummyLinearModel(BaseEstimator):
    # Does not natively support multiple targets
    def fit(self, x, y):
        self.coef_ = np.ones(x.shape[1])
        self.intercept_ = 0
        return self

    def predict(self, x):
        return x


class DummyEmptyModel(BaseEstimator):
    # Does not have fit or predict methods
    def __init__(self):
        self.fit_intercept = False
        self.normalize_columns = False


class DummyModelNoCoef(BaseEstimator):
    # Does not set the coef_ attribute
    def fit(self, x, y):
        self.intercept_ = 0
        return self

    def predict(self, x):
        return x


def _align_optimizer_and_1dfeatures(
    opt: BaseOptimizer, features: NDArray
) -> tuple[BaseOptimizer, NDArray]:
    # This is a hack until constraints are moved from init to fit
    if isinstance(opt, TrappingSR3):
        opt = TrappingSR3(_n_tgts=1, _include_bias=False)
        features = np.hstack([features, features])
    elif isinstance(opt, SSR):
        features = np.hstack([features, features])
    else:
        features = features
    return opt, features


@pytest.mark.parametrize(
    "cls, support",
    [
        (Lasso, True),
        (STLSQ, True),
        (SSR, True),
        (FROLS, True),
        (SR3, True),
        (ConstrainedSR3, True),
        (StableLinearSR3, True),
        (TrappingSR3, True),
        (DummyLinearModel, False),
        (SBR, True),
    ],
)
def test_supports_multiple_targets(cls, support):
    assert supports_multiple_targets(cls()) == support


@pytest.fixture(params=["data_derivative_1d", "data_derivative_2d"])
def data(request):
    return request.getfixturevalue(request.param)


@pytest.mark.parametrize(
    "optimizer",
    [
        STLSQ(),
        SSR(),
        SSR(criteria="model_residual"),
        FROLS(),
        SR3(),
        ConstrainedSR3(),
        StableLinearSR3(),
        TrappingSR3(_n_tgts=1),
        Lasso(fit_intercept=False),
        ElasticNet(fit_intercept=False),
        DummyLinearModel(),
        MIOSR(),
        SBR(num_warmup=10, num_samples=10),
    ],
    ids=lambda param: type(param),
)
def test_fit(data_derivative_1d, optimizer, set_jax_pcc_env):
    x, x_dot = data_derivative_1d
    if len(x.shape) == 1:
        x = x.reshape(-1, 1)
    optimizer, x = _align_optimizer_and_1dfeatures(optimizer, x)
    opt = WrappedOptimizer(optimizer, unbias=False)
    opt.fit(x, x_dot)

    check_is_fitted(opt)
    assert opt.complexity >= 0
    if len(x_dot.shape) > 1:
        assert opt.coef_.shape == (x.shape[1], x_dot.shape[1])
    else:
        assert opt.coef_.shape == (1, x.shape[1])


@pytest.mark.parametrize(
    "optimizer",
    [STLSQ(), SSR(), SSR(criteria="model_residual"), FROLS(), SR3(), MIOSR(), SBR()],
)
def test_not_fitted(optimizer):
    with pytest.raises(NotFittedError):
        optimizer.predict(np.ones((1, 3)))


@pytest.mark.parametrize(
    "optimizer",
    [
        STLSQ(),
        SR3(),
        SBR(
            num_warmup=1,
            num_samples=1,
            nuts_kwargs={"max_tree_depth": 1, "target_accept_prob": 0.1},
        ),
    ],
    ids=type,
)
def test_complexity_not_fitted(optimizer, data_derivative_2d, set_jax_pcc_env):
    with pytest.raises(NotFittedError):
        optimizer.complexity

    x, _ = data_derivative_2d
    optimizer.fit(x, x)
    assert optimizer.complexity > 0


@pytest.mark.parametrize("kwargs", [{"normalize_columns": True}, {"copy_X": False}])
def test_alternate_parameters(data_derivative_1d, kwargs):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)

    model = STLSQ(**kwargs)
    model.fit(x, x_dot)
    model.fit(x, x_dot, sample_weight=x[:, 0])

    check_is_fitted(model)


@pytest.mark.parametrize(
    "optimizer",
    [
        STLSQ,
        SSR,
        FROLS,
        SR3,
        ConstrainedSR3,
        StableLinearSR3,
        TrappingSR3,
        MIOSR,
    ],
)
def test_sample_weight_optimizers(data_1d, optimizer):
    y, t = data_1d
    opt = optimizer()
    opt, x = _align_optimizer_and_1dfeatures(opt, y)
    sample_weight = np.ones(x[:, 0].shape)
    sample_weight[::2] = 0
    opt.fit(x, y, sample_weight=sample_weight)
    check_is_fitted(opt)


@pytest.mark.parametrize("params", [dict(threshold=-1), dict(max_iter=0)])
def test_STLSQ_bad_parameters(params):
    with pytest.raises(ValueError):
        STLSQ(**params)


@pytest.mark.parametrize(
    "optimizer", [SR3, ConstrainedSR3, StableLinearSR3, TrappingSR3]
)
@pytest.mark.parametrize("params", [dict(reg_weight_lam=-1), dict(max_iter=0)])
def test_general_bad_parameters(optimizer, params):
    with pytest.raises(ValueError):
        optimizer(**params)


@pytest.mark.parametrize("optimizer", [SR3, ConstrainedSR3, StableLinearSR3])
@pytest.mark.parametrize(
    "params",
    [
        dict(relax_coeff_nu=0),
        dict(tol=0),
        dict(trimming_fraction=-1),
        dict(trimming_fraction=2),
    ],
)
def test_sr3_bad_parameters(optimizer, params):
    with pytest.raises(ValueError):
        optimizer(**params)


@pytest.mark.parametrize(
    "params",
    [
        dict(eta=-1),
        dict(tol=0),
        dict(tol_m=0),
        dict(eps_solver=0),
        dict(eta=1, alpha_m=-1),
        dict(eta=1, alpha_A=-1),
        dict(gamma=1),
        dict(regularizer="l0"),
        dict(max_iter=0),
        dict(eta=10, alpha_m=20),
        dict(eta=10, alpha_A=20),
        dict(
            constraint_lhs=np.zeros((10, 10)),
            constraint_rhs=np.zeros(10),
            constraint_order="None",
        ),
    ],
)
def test_trapping_bad_parameters(params):
    with pytest.raises(ValueError):
        TrappingSR3(_n_tgts=1, **params)


def test_trapping_objective_print():
    # test error in verbose print logic when max_iter < 10
    opt = TrappingSR3(_n_tgts=1, max_iter=2, verbose=True)
    arr = np.ones((1, 1))
    opt.PQ_ = np.ones((1, 1, 1, 1, 1))
    opt._objective(arr, arr, arr, arr, arr, 1)


@pytest.mark.parametrize(
    "params",
    [
        dict(tol=0),
        dict(max_iter=-1),
        dict(regularizer="l0"),
        dict(reg_weight_lam=-1),
        dict(regularizer="weighted_l1", reg_weight_lam=1),
        dict(regularizer="weighted_l1"),
        dict(model_subset=0),
        dict(model_subset=[50]),
        dict(model_subset=[0, 0.5, 1]),
    ],
)
def test_sindypi_bad_parameters(data_lorenz, params):
    x, t = data_lorenz
    with pytest.raises(ValueError):
        opt = SINDyPI(**params)
        model = SINDy(optimizer=opt)
        model.fit(x, t=t)


@pytest.mark.parametrize(
    "params",
    [
        dict(tol=1e-3),
        dict(regularizer="l1"),
        dict(regularizer="weighted_l1", reg_weight_lam=np.zeros((10, 10))),
        dict(regularizer="l2"),
        dict(regularizer="weighted_l2", reg_weight_lam=np.zeros((10, 10))),
        dict(model_subset=[5]),
    ],
)
def test_sindypi_fit(params):
    dt = 0.01
    T = 5
    t = np.arange(0, T + dt, dt)
    x0_train = [0.55]
    x_train = solve_ivp(enzyme, (t[0], t[-1]), x0_train, t_eval=t).y.T

    # initialize a quartic polynomial library for x
    x_library_functions = [
        lambda x: x,
        lambda x, y: x * y,
        lambda x: x**2,
        lambda x, y, z: x * y * z,
        lambda x, y: x * y**2,
        lambda x: x**3,
        lambda x, y, z, w: x * y * z * w,
        lambda x, y, z: x * y * z**2,
        lambda x, y: x * y**3,
        lambda x: x**4,
    ]
    # initialize a linear polynomial library for x_dot
    x_dot_library_functions = [lambda x: x]

    # library function names includes both the x_library_functions
    # and x_dot_library_functions names
    library_function_names = [
        lambda x: x,
        lambda x, y: x + y,
        lambda x: x + x,
        lambda x, y, z: x + y + z,
        lambda x, y: x + y + y,
        lambda x: x + x + x,
        lambda x, y, z, w: x + y + z + w,
        lambda x, y, z: x + y + z + z,
        lambda x, y: x + y + y + y,
        lambda x: x + x + x + x,
        lambda x: x,
    ]

    # Need to pass time base to the library so can build the x_dot library from x
    sindy_library = SINDyPILibrary(
        library_functions=x_library_functions,
        x_dot_library_functions=x_dot_library_functions,
        t=t,
        function_names=library_function_names,
        include_bias=True,
    )

    opt = SINDyPI(**params)
    model = SINDy(
        optimizer=opt,
        feature_library=sindy_library,
        differentiation_method=FiniteDifference(drop_endpoints=True),
    )
    model.fit(x_train, t=t)
    assert np.shape(opt.coef_) == (10, 10)


@pytest.mark.parametrize(
    "params",
    [
        dict(sparsity_coef_tau0=0),
        dict(sparsity_coef_tau0=-1),
        dict(slab_shape_nu=0),
        dict(slab_shape_nu=-1),
        dict(slab_shape_s=0),
        dict(slab_shape_s=-1),
        dict(noise_hyper_lambda=0),
        dict(noise_hyper_lambda=-1),
        dict(num_warmup=0.5),
        dict(num_warmup=-1),
        dict(num_samples=0.5),
        dict(num_samples=-1),
    ],
)
def test_sbr_bad_parameters(params):
    with pytest.raises(ValueError):
        SBR(**params)


def test_sbr_accurate(set_jax_pcc_env):
    # It's really hard to tune SBR to get desired shrinkage
    # This just tests that SBR fits "close" to unregularized regression
    x = np.tile(np.eye(2), 4).reshape((-1, 2))
    y = np.tile([[1], [1e-1]], 4).reshape((-1, 1))
    opt = SBR(num_warmup=50, num_samples=50).fit(x, y)
    result = opt.coef_
    unregularized = np.array([[1, 1e-1]])
    np.testing.assert_allclose(result, unregularized, atol=1e-3)
    assert hasattr(opt, "mcmc_")
    expected_names = ["beta", "c_sq", "lambda", "sigma", "tau"]
    result_names = opt.mcmc_.get_samples().keys()
    assert all(expected in result_names for expected in expected_names)


@pytest.mark.parametrize(
    "params",
    [
        dict(regularizer="l1", reg_weight_lam=0),
        dict(regularizer="l1", reg_weight_lam=1e-5),
        dict(regularizer="weighted_l1", reg_weight_lam=np.zeros((3, 9))),
        dict(regularizer="weighted_l1", reg_weight_lam=1e-5 * np.ones((3, 9))),
        dict(regularizer="l2", reg_weight_lam=0),
        dict(regularizer="l2", reg_weight_lam=1e-5),
        dict(regularizer="weighted_l2", reg_weight_lam=np.zeros((3, 9))),
        dict(regularizer="weighted_l2", reg_weight_lam=1e-5 * np.ones((3, 9))),
    ],
)
def test_sr3_quadratic_library(params):
    x = np.random.standard_normal((100, 3))
    library_functions = [
        lambda x: x,
        lambda x, y: x * y,
        lambda x: x**2,
    ]
    library_function_names = [
        lambda x: str(x),
        lambda x, y: "{} * {}".format(x, y),
        lambda x: "{}^2".format(x),
    ]
    sindy_library = CustomLibrary(
        library_functions=library_functions, function_names=library_function_names
    )

    # Test SR3
    opt = SR3(**params)
    model = SINDy(optimizer=opt, feature_library=sindy_library)
    model.fit(x, t=1)
    check_is_fitted(model)


@pytest.mark.parametrize(
    "params",
    [
        dict(regularizer="l1", reg_weight_lam=0),
        dict(regularizer="l1", reg_weight_lam=1e-5),
        dict(regularizer="weighted_l1", reg_weight_lam=np.zeros((3, 9))),
        dict(regularizer="weighted_l1", reg_weight_lam=1e-5 * np.ones((3, 9))),
        dict(regularizer="l2", reg_weight_lam=0),
        dict(regularizer="l2", reg_weight_lam=1e-5),
        dict(regularizer="weighted_l2", reg_weight_lam=np.zeros((3, 9))),
        dict(regularizer="weighted_l2", reg_weight_lam=1e-5 * np.ones((3, 9))),
    ],
)
def test_constrained_sr3_quadratic_library(params):
    x = np.random.standard_normal((100, 3))
    library_functions = [
        lambda x: x,
        lambda x, y: x * y,
        lambda x: x**2,
    ]
    library_function_names = [
        lambda x: str(x),
        lambda x, y: "{} * {}".format(x, y),
        lambda x: "{}^2".format(x),
    ]
    sindy_library = CustomLibrary(
        library_functions=library_functions, function_names=library_function_names
    )

    # Test constrained SR3 without constraints
    opt = ConstrainedSR3(**params)
    model = SINDy(optimizer=opt, feature_library=sindy_library)
    model.fit(x, t=1)
    check_is_fitted(model)

    # rerun with identity constraints
    r = 3
    N = 9
    p = r + r * (r - 1) + int(r * (r - 1) * (r - 2) / 6.0)
    constraint_rhs = np.zeros(p)
    constraint_matrix = np.eye(p, r * N)

    # Test constrained SR3 with constraints
    opt = ConstrainedSR3(
        constraint_lhs=constraint_matrix, constraint_rhs=constraint_rhs, **params
    )
    model = SINDy(optimizer=opt, feature_library=sindy_library)
    model.fit(x, t=1)
    check_is_fitted(model)
    assert np.allclose((model.coefficients().flatten())[:p], 0.0)


@pytest.mark.parametrize(
    ["regularizer", "lam", "expected"],
    [
        ("l1", np.array([[2]]), 20),
        ("weighted_l1", np.array([[3, 2, 0.5]]).T, 14.5),
        ("l2", np.array([[2]]), 76),
        ("weighted_l2", np.array([[3, 2, 0.5]]).T, 42.5),
    ],
)
def test_constrained_sr3_penalty_term(regularizer, lam, expected):
    xi = cp.Variable(3)
    penalty = ConstrainedSR3._calculate_penalty(regularizer, np.ravel(lam), xi)
    xi.value = np.array([-2, 3, 5])
    np.testing.assert_allclose(penalty.value, expected)


def test_stable_linear_sr3_linear_library():
    x = np.ones((2, 1))
    opt = StableLinearSR3()
    opt.fit(x, x)
    check_is_fitted(opt)

    constraint_rhs = np.zeros((1, 1))
    constraint_matrix = np.eye(1)
    opt = StableLinearSR3(
        constraint_lhs=constraint_matrix, constraint_rhs=constraint_rhs
    )
    opt.fit(x, x)
    check_is_fitted(opt)
    assert np.allclose(opt.coef_.flatten(), 0.0)


@pytest.mark.parametrize("bias", (True, False))
@pytest.mark.parametrize("method", ("global", "local"))
@pytest.mark.parametrize("reg_weight", (0.0, 1e-1))
def test_trapping_sr3_quadratic_library(bias, method, reg_weight):
    t = np.arange(0, 1, 0.1)
    x = np.exp(-t).reshape((-1, 1))
    x_dot = -x
    features = np.hstack([x, x**2])
    if bias:
        features = np.hstack([np.ones_like(x), features])

    params = {
        "_n_tgts": 1,
        "_include_bias": bias,
        "method": method,
        "reg_weight_lam": reg_weight,
    }

    opt = TrappingSR3(**params)
    opt.fit(features, x_dot)
    check_is_fitted(opt)

    # Rerun with identity constraints
    r = x.shape[1]
    N = 2 + bias
    params["constraint_rhs"] = np.zeros(r * N)
    params["constraint_lhs"] = np.eye(r * N, r * N)

    opt = TrappingSR3(**params)
    opt.fit(features, x_dot)
    check_is_fitted(opt)
    # check if solve was infeasible first
    if not np.allclose(opt.m_history_[-1], opt.m_history_[0]):
        assert np.allclose((opt.coef_.flatten())[0], 0.0, atol=1e-5)


@pytest.mark.parametrize(
    "error, optimizer, params",
    [
        (ValueError, STLSQ, dict(alpha=-1)),
        (ValueError, SSR, dict(alpha=-1)),
        (ValueError, SSR, dict(criteria="None")),
        (ValueError, SSR, dict(max_iter=-1)),
        (ValueError, FROLS, dict(max_iter=-1)),
        (NotImplementedError, SR3, dict(regularizer="l3")),
        (NotImplementedError, ConstrainedSR3, dict(regularizer="l3")),
        (NotImplementedError, StableLinearSR3, dict(regularizer="l3")),
        (
            ValueError,
            ConstrainedSR3,
            dict(
                inequality_constraints=True,
                constraint_lhs=np.zeros((1, 1)),
                constraint_rhs=np.zeros(1),
                regularizer="l0",
            ),
        ),
        (ValueError, ConstrainedSR3, dict(inequality_constraints=True)),
        (ValueError, SR3, dict(regularizer="weighted_l0", reg_weight_lam=None)),
        (ValueError, SR3, dict(regularizer="weighted_l1", reg_weight_lam=None)),
        (ValueError, SR3, dict(regularizer="weighted_l2", reg_weight_lam=None)),
        (ValueError, SR3, dict(reg_weight_lam=-np.ones((5, 5)))),
        (ValueError, SR3, dict(initial_guess=np.zeros(3))),
        (
            ValueError,
            ConstrainedSR3,
            dict(regularizer="weighted_l0", reg_weight_lam=None),
        ),
        (
            ValueError,
            ConstrainedSR3,
            dict(regularizer="weighted_l1", reg_weight_lam=None),
        ),
        (
            ValueError,
            ConstrainedSR3,
            dict(regularizer="weighted_l2", reg_weight_lam=None),
        ),
        (ValueError, ConstrainedSR3, dict(reg_weight_lam=-np.ones((5, 5)))),
        (ValueError, ConstrainedSR3, dict(initial_guess=np.zeros(3))),
        (
            ValueError,
            ConstrainedSR3,
            dict(
                constraint_lhs=np.zeros((10, 10)),
                constraint_rhs=np.zeros(10),
                constraint_order="None",
            ),
        ),
    ],
)
def test_specific_bad_parameters(error, optimizer, params, data_lorenz):
    x, t = data_lorenz
    with pytest.raises(error):
        opt = optimizer(**params)
        model = SINDy(optimizer=opt)
        model.fit(x, t=t)


def test_bad_optimizers(data_derivative_1d):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)
    x_dot = x_dot.reshape(-1, 1)
    with pytest.raises(InvalidParameterError):
        # Error: optimizer does not have a callable fit method
        opt = WrappedOptimizer(DummyEmptyModel())
        opt.fit(x, x_dot)

    with pytest.raises(AttributeError):
        # Error: object has no attribute 'coef_'
        opt = WrappedOptimizer(DummyModelNoCoef())
        opt.fit(x, x_dot)


@pytest.mark.parametrize("optimizer", [SR3, ConstrainedSR3])
def test_initial_guess_sr3(optimizer):
    x = np.random.standard_normal((10, 3))
    x_dot = np.random.standard_normal((10, 2))

    control_model = optimizer(max_iter=1).fit(x, x_dot)

    initial_guess = np.random.standard_normal((x_dot.shape[1], x.shape[1]))
    guess_model = optimizer(max_iter=1, initial_guess=initial_guess).fit(x, x_dot)
    assert np.any(np.not_equal(control_model.coef_, guess_model.coef_))


# The different capitalizations are intentional;
# I want to make sure different versions are recognized
@pytest.mark.parametrize("optimizer", [SR3, ConstrainedSR3])
@pytest.mark.parametrize("regularizer", ["L0", "l1"])
def test_prox_functions(data_derivative_1d, optimizer, regularizer):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)
    model = optimizer(regularizer=regularizer)
    model.fit(x, x_dot)
    check_is_fitted(model)


@pytest.mark.parametrize("regularizer", ["weighted_l0", "weighted_l1"])
def test_weighted_prox_functions(data, regularizer):
    x, x_dot = data
    if x.ndim == 1:
        x = x.reshape(-1, 1)
        reg_weight_lam = np.ones((1, 1))
    else:
        reg_weight_lam = np.ones((x_dot.shape[1], x.shape[1]))

    model = ConstrainedSR3(regularizer=regularizer, reg_weight_lam=reg_weight_lam)
    model.fit(x, x_dot)
    check_is_fitted(model)


@pytest.mark.parametrize("regularizer", ["L0", "l1"])
def test_constrained_sr3_prox_functions(data_derivative_1d, regularizer):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)
    model = ConstrainedSR3(regularizer=regularizer)
    model.fit(x, x_dot)
    check_is_fitted(model)


@pytest.mark.parametrize(
    ("opt_cls", "opt_args"),
    (
        (SR3, {"trimming_fraction": 0.1}),
        (ConstrainedSR3, {"constraint_lhs": [1], "constraint_rhs": [1]}),
        (ConstrainedSR3, {"trimming_fraction": 0.1}),
        (TrappingSR3, {"_n_tgts": 1, "constraint_lhs": [1], "constraint_rhs": [1]}),
        (StableLinearSR3, {"constraint_lhs": [1], "constraint_rhs": [1]}),
        (StableLinearSR3, {"trimming_fraction": 0.1}),
        (SINDyPI, {}),
        (MIOSR, {"constraint_lhs": [1]}),
        (SBR, {}),
    ),
)
def test_illegal_unbias(data_derivative_1d, opt_cls, opt_args):
    x, x_dot = data_derivative_1d
    with pytest.raises(ValueError):
        opt_cls(unbias=True, **opt_args).fit(x, x_dot)


def test_unbias(data_derivative_1d):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)
    x_dot = x_dot.reshape(-1, 1)

    optimizer_biased = STLSQ(threshold=0.01, alpha=0.1, max_iter=1, unbias=False)
    optimizer_biased.fit(x, x_dot)

    optimizer_unbiased = STLSQ(threshold=0.01, alpha=0.1, max_iter=1, unbias=True)
    optimizer_unbiased.fit(x, x_dot)

    assert (
        norm(optimizer_biased.coef_ - optimizer_unbiased.coef_)
        / norm(optimizer_unbiased.coef_)
        > 1e-9
    )


def test_unbias_external(data_derivative_1d):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)
    x_dot = x_dot.reshape(-1, 1)

    optimizer_biased = WrappedOptimizer(
        Lasso(alpha=0.1, fit_intercept=False, max_iter=1), unbias=False
    )
    optimizer_biased.fit(x, x_dot)

    optimizer_unbiased = WrappedOptimizer(
        Lasso(alpha=0.1, fit_intercept=False, max_iter=1), unbias=True
    )
    optimizer_unbiased.fit(x, x_dot)

    assert (
        norm(optimizer_biased.coef_ - optimizer_unbiased.coef_)
        / (norm(optimizer_unbiased.coef_) + 1e-5)
        > 1e-9
    )


@pytest.mark.parametrize("OptCls", [SR3, ConstrainedSR3])
def test_sr3_trimming(OptCls, data_linear_oscillator_corrupted):
    X, X_dot, trimming_array = data_linear_oscillator_corrupted

    optimizer_without_trimming = OptCls(unbias=False)
    optimizer_without_trimming.fit(X, X_dot)

    optimizer_trimming = OptCls(trimming_fraction=0.15, unbias=False)
    optimizer_trimming.fit(X, X_dot)

    # Check that trimming found the right samples to remove
    np.testing.assert_array_equal(optimizer_trimming.trimming_array, trimming_array)

    # Check that the coefficients found by the optimizer with trimming
    # are closer to the true coefficients than the coefficients found by the
    # optimizer without trimming
    true_coef = np.array([[-2.0, 0.0], [0.0, 1.0]])
    assert norm(true_coef - optimizer_trimming.coef_) < norm(
        true_coef - optimizer_without_trimming.coef_
    )


@pytest.mark.parametrize("optimizer", [SR3, ConstrainedSR3])
def test_sr3_disable_trimming(optimizer, data_linear_oscillator_corrupted):
    x, x_dot, _ = data_linear_oscillator_corrupted

    model_plain = optimizer()
    model_plain.fit(x, x_dot)

    model_trimming = optimizer(trimming_fraction=0.5)
    model_trimming.disable_trimming()
    model_trimming.fit(x, x_dot)

    np.testing.assert_allclose(model_plain.coef_, model_trimming.coef_)


@pytest.mark.parametrize("optimizer", [SR3, ConstrainedSR3])
def test_sr3_enable_trimming(optimizer, data_linear_oscillator_corrupted):
    x, x_dot, _ = data_linear_oscillator_corrupted

    model_plain = optimizer()
    model_plain.enable_trimming(trimming_fraction=0.5)
    model_plain.fit(x, x_dot)

    model_trimming = optimizer(trimming_fraction=0.5)
    model_trimming.fit(x, x_dot)

    np.testing.assert_allclose(model_plain.coef_, model_trimming.coef_)


@pytest.mark.parametrize(
    "optimizer",
    [
        STLSQ(max_iter=1),
        SR3(max_iter=1),
        ConstrainedSR3(max_iter=1),
        StableLinearSR3(max_iter=1),
        TrappingSR3(_n_tgts=1, max_iter=1),
    ],
)
def test_fit_warn(data_derivative_1d, optimizer):
    x, x_dot = data_derivative_1d
    x = x.reshape(-1, 1)
    optimizer.max_iter = 0  # normally prohibited in constructor

    with pytest.warns(ConvergenceWarning):
        optimizer.fit(x, x_dot)


@pytest.mark.parametrize(
    "optimizer",
    [
        (ConstrainedSR3, {"max_iter": 80}),
        (TrappingSR3, {"_n_tgts": 3, "max_iter": 100, "eps_solver": 1e-5}),
        (MIOSR, {}),
    ],
    ids=lambda param: param[0].__name__ + " " + ",".join([key for key in param[1]]),
)
@pytest.mark.parametrize("target_value", [0, -1, 3])
def test_feature_format_constraints(data_linear_combination, optimizer, target_value):
    # Solution is x_dot = x.dot(np.array([[1, 1, 0], [0, 1, 1]]))
    x, y = data_linear_combination

    constraint_rhs = target_value * np.ones(2)
    constraint_lhs = np.zeros((2, x.shape[1], y.shape[1]))

    # Should force corresponding entries of coef_ to be target_value
    constraint_lhs[0, 1, 1] = 1
    constraint_lhs[1, 2, 2] = 1
    # reshape to "feature" order
    constraint_lhs = np.reshape(constraint_lhs, (constraint_lhs.shape[0], -1))

    model = optimizer[0](
        constraint_lhs=constraint_lhs,
        constraint_rhs=constraint_rhs,
        constraint_order="feature",
        **optimizer[1],
    )
    model.fit(x, y)

    np.testing.assert_allclose(
        np.array([model.coef_[1, 1], model.coef_[2, 2]]), target_value, atol=1e-7
    )


@pytest.mark.parametrize(
    "optimizer",
    [
        (ConstrainedSR3, {"max_iter": 80}),
        (StableLinearSR3, {}),
        (TrappingSR3, {"_n_tgts": 3, "max_iter": 200, "eps_solver": 1e-5}),
        (MIOSR, {}),
    ],
    ids=lambda param: param[0].__name__ + " " + ",".join([key for key in param[1]]),
)
@pytest.mark.parametrize("target_value", [0, -1, 3])
def test_target_format_constraints(data_linear_combination, optimizer, target_value):
    x, y = data_linear_combination

    constraint_rhs = target_value * np.ones(2)
    constraint_lhs = np.zeros((2, x.shape[1], y.shape[1]))

    # Should force corresponding entries of coef_ to be target_value
    constraint_lhs[0, 2, 1] = 1
    constraint_lhs[1, 1, 2] = 1
    # reshape to "target" order
    constraint_lhs = np.reshape(
        np.transpose(constraint_lhs, [0, 2, 1]), (constraint_lhs.shape[0], -1)
    )

    model = optimizer[0](
        constraint_lhs=constraint_lhs,
        constraint_rhs=constraint_rhs,
        constraint_order="target",
        **optimizer[1],
    )
    model.fit(x, y)

    np.testing.assert_allclose(
        np.array([model.coef_[1, 2], model.coef_[2, 1]]), target_value, atol=1e-7
    )


@pytest.mark.parametrize(
    "params",
    [
        dict(regularizer="l1", reg_weight_lam=0.0005),
        dict(regularizer="weighted_l1", reg_weight_lam=0.0005 * np.ones((3, 10))),
        dict(regularizer="l2", reg_weight_lam=0.0005),
        dict(regularizer="weighted_l2", reg_weight_lam=0.0005 * np.ones((3, 10))),
    ],
)
def test_constrained_inequality_constraints(data_lorenz, params):
    x, t = data_lorenz
    constraint_rhs = np.array([-10.0, 28.0])
    constraint_matrix = np.zeros((2, 30))
    constraint_matrix[0, 1] = 1.0
    constraint_matrix[1, 11] = 1.0
    feature_names = ["x", "y", "z"]

    poly_lib = PolynomialLibrary(degree=2)
    # Run constrained SR3
    opt = ConstrainedSR3(
        constraint_lhs=constraint_matrix,
        constraint_rhs=constraint_rhs,
        constraint_order="feature",
        inequality_constraints=True,
        **params,
    )
    model = SINDy(
        optimizer=opt,
        feature_library=poly_lib,
        differentiation_method=FiniteDifference(drop_endpoints=True),
    )
    model.fit(
        x,
        t=t[1] - t[0],
        feature_names=feature_names,
    )
    # This sometimes fails with L2 norm so just check the model is fitted
    check_is_fitted(model)


@pytest.mark.parametrize(
    "params",
    [
        dict(regularizer="l1", reg_weight_lam=2, expected=2.5),
        dict(
            regularizer="weighted_l1",
            reg_weight_lam=0.5 * np.ones((1, 2)),
            expected=1.0,
        ),
        dict(regularizer="l2", reg_weight_lam=2, expected=1.5),
        dict(
            regularizer="weighted_l2",
            reg_weight_lam=0.5 * np.ones((1, 2)),
            expected=0.75,
        ),
        dict(regularizer="l1", reg_weight_lam=0, expected=0.5),
        dict(
            regularizer="weighted_l1",
            reg_weight_lam=0.0 * np.ones((1, 2)),
            expected=0.5,
        ),
        dict(regularizer="l2", reg_weight_lam=0.0, expected=0.5),
        dict(
            regularizer="weighted_l2",
            reg_weight_lam=0.0 * np.ones((1, 2)),
            expected=0.5,
        ),
    ],
    ids=lambda d: d["regularizer"],
)
def test_trapping_cost_function(params):
    # TODO: are all these parameters necessary?  What are we testing?
    expected = params.pop("expected")
    opt = TrappingSR3(**params)
    x = np.eye(2)
    y = np.ones(2)
    xi, cost = opt._create_var_and_part_cost(2, x, y)
    xi.value = np.array([0.5, 0.5])
    np.testing.assert_allclose(cost.value, expected)


def test_trapping_inequality_constraints():
    t = np.arange(0, 1, 0.1)
    x = np.stack((t, t**2)).T
    y = x[:, 0] + 0.1 * x[:, 1]
    constraint_rhs = np.array([0.1])
    constraint_matrix = np.zeros((1, 2))
    constraint_matrix[0, 1] = 0.1

    # Run Trapping SR3
    opt = TrappingSR3(
        constraint_lhs=constraint_matrix,
        constraint_rhs=constraint_rhs,
        constraint_order="feature",
        inequality_constraints=True,
    )
    opt.fit(x, y)
    assert np.all(np.dot(constraint_matrix, (opt.coef_).flatten()) <= constraint_rhs)


@pytest.mark.parametrize(
    "params",
    [
        dict(target_sparsity=2),
        dict(target_sparsity=7),
    ],
)
def test_miosr_equality_constraints(data_lorenz, params):
    x, t = data_lorenz
    constraint_rhs = np.array([-10.0, 28.0])
    constraint_matrix = np.zeros((2, 30))
    constraint_matrix[0, 1] = 1.0
    constraint_matrix[1, 11] = 1.0
    feature_names = ["x", "y", "z"]

    opt = MIOSR(
        constraint_lhs=constraint_matrix,
        constraint_rhs=constraint_rhs,
        constraint_order="target",
        **params,
    )
    poly_lib = PolynomialLibrary(degree=2)
    model = SINDy(
        optimizer=opt,
        feature_library=poly_lib,
    )
    model.fit(
        x,
        t=t[1] - t[0],
        feature_names=feature_names,
    )
    assert np.allclose(
        np.dot(constraint_matrix, (model.coefficients()).flatten()),
        constraint_rhs,
        atol=1e-3,
    )


def test_inequality_constraints_reqs():
    constraint_rhs = np.array([-10.0, -2.0])
    constraint_matrix = np.zeros((2, 30))
    constraint_matrix[0, 6] = 1.0
    constraint_matrix[1, 17] = 1.0
    with pytest.raises(ValueError):
        TrappingSR3(
            reg_weight_lam=0.0,
            constraint_lhs=constraint_matrix,
            constraint_rhs=constraint_rhs,
            constraint_order="feature",
            inequality_constraints=True,
        )


def test_normalize_columns(data_derivative_1d):
    x, x_dot = data_derivative_1d
    x = np.reshape(x, (-1, 1))
    x_dot = np.reshape(x_dot, (-1, 1))
    cols = np.hstack((x, x_dot))
    norm, ncols = _normalize_features(cols)
    result = np.linalg.norm(ncols, axis=0)
    expected = [1.0, 1.0]
    np.testing.assert_allclose(result, expected)
    np.testing.assert_allclose(ncols * norm, cols)


@pytest.mark.parametrize(
    "optimizer_params",
    (
        {"library_ensemble": True, "n_models": 2},
        {"bagging": True, "n_models": 2, "n_subset": 2},
        {"library_ensemble": True, "bagging": True, "n_models": 2, "n_subset": 2},
    ),
)
def test_ensemble_optimizer(data_lorenz, optimizer_params):
    x, t = data_lorenz
    optimizer = EnsembleOptimizer(STLSQ(), **optimizer_params)
    optimizer.fit(x, x)
    assert optimizer.coef_.shape == (3, 3)
    assert len(optimizer.coef_list) == 2


@pytest.mark.parametrize(
    "params",
    [
        dict(),
        dict(bagging=True, n_models=0),
        dict(bagging=True, n_subset=0),
        dict(library_ensemble=True, n_candidates_to_drop=0),
    ],
)
def test_bad_ensemble_params(data_lorenz, params):
    with pytest.raises(ValueError):
        EnsembleOptimizer(opt=STLSQ(), **params)


def test_ssr_criteria(data_lorenz):
    x, t = data_lorenz
    opt = SSR(normalize_columns=True, criteria="model_residual", kappa=1e-3)
    model = SINDy(optimizer=opt)
    model.fit(x, t)
    assert np.shape(opt.coef_) == (3, 10)


@pytest.mark.parametrize(
    "optimizer",
    [
        STLSQ,
        SSR,
        FROLS,
        SR3,
        ConstrainedSR3,
        StableLinearSR3,
        TrappingSR3,
        MIOSR,
    ],
)
def test_optimizers_verbose(data_1d, optimizer):
    y, _ = data_1d
    opt = optimizer(verbose=True)
    opt, x = _align_optimizer_and_1dfeatures(opt, y)
    opt.verbose = True
    opt.fit(x, y)
    check_is_fitted(opt)


@pytest.mark.parametrize(
    "optimizer",
    [
        SINDyPI,
        ConstrainedSR3,
        StableLinearSR3,
        TrappingSR3,
    ],
)
def test_optimizers_verbose_cvxpy(data_1d, optimizer):
    y, _ = data_1d
    opt = optimizer(verbose_cvxpy=True)
    opt, x = _align_optimizer_and_1dfeatures(opt, y)
    opt.fit(x, y)
    check_is_fitted(opt)


def test_frols_error_linear_dependence():
    opt = FROLS(normalize_columns=True)
    x = np.array([[1.0, 1.0]])
    y = np.array([[1.0, 1.0]])
    with pytest.raises(ValueError):
        opt.fit(x, y)


def test_sparse_subset_multitarget():
    A = np.eye(4)
    b = np.array([[1, 1, 0.5, 1], [1, 1, 1, 0.5]]).T
    opt = STLSQ(unbias=False, threshold=0.5, alpha=0.1, sparse_ind=[2, 3])
    opt.fit(A, b)
    X = opt.coef_
    Y = opt.optvar_non_sparse_
    assert X[0, 0] == 0.0
    assert 0.0 < X[0, 1] < 1.0
    np.testing.assert_equal(Y[:, :2], np.ones((2, 2)))
    assert X[1, 1] == 0.0
    assert 0.0 < X[1, 0] < 1.0


def test_sparse_subset_off_diagonal():
    A = np.array([[1, 1], [0, 1]])
    b = np.array([1, 1])
    opt = STLSQ(unbias=False, threshold=0.1, alpha=0.1, sparse_ind=[1])
    opt.fit(A, b)
    X = opt.coef_
    Y = opt.optvar_non_sparse_
    assert Y[0, 0] > 0.0 and Y[0, 0] < 0.5
    assert X[0, 0] > 0.5 and X[0, 0] < 1.0


def test_sparse_subset_unbias():
    A = np.array([[1, 1], [0, 1]])
    b = np.array([1, 1])
    opt = STLSQ(unbias=True, threshold=0.1, alpha=0.1, sparse_ind=[1])
    opt.fit(A, b)
    X = opt.coef_
    Y = opt.optvar_non_sparse_
    assert np.abs(Y[0, 0]) < 2e-16
    assert np.abs(X[0, 0] - 1.0) < 2e-16


def test_remove_and_decrement():
    existing_vals = np.array([2, 3, 4, 5])
    vals_to_remove = np.array([3, 5])
    expected = np.array([2, 3])
    result = _remove_and_decrement(
        existing_vals=existing_vals, vals_to_remove=vals_to_remove
    )
    np.testing.assert_array_equal(expected, result)


@pytest.mark.parametrize(
    ("opt_cls", "opt_args"),
    (
        (MIOSR, {"target_sparsity": 7}),
        (SBR, {"num_warmup": 10, "num_samples": 10}),
        (SR3, {}),
        (TrappingSR3, {"_n_tgts": 3, "_include_bias": True}),
    ),
)
def test_pickle(data_lorenz, opt_cls, opt_args, set_jax_pcc_env):
    x, t = data_lorenz
    y = PolynomialLibrary(degree=2).fit_transform(x)
    opt = opt_cls(**opt_args).fit(y, x)
    expected = opt.coef_
    new_opt = pickle.loads(pickle.dumps(opt))
    result = new_opt.coef_
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("kappa", (None, 0.1), ids=["inflection", "L0"])
def test_ssr_history_selection(kappa):
    rng = np.random.default_rng(1)
    x = rng.normal(size=(30, 8))
    expected = np.array([[1, 1, 1, 0, 0, 0, 0, 0]])
    y = x @ expected.T

    x += np.random.normal(size=(30, 8), scale=1e-2)
    opt = SSR(kappa=kappa)
    result = opt.fit(x, y).coef_

    assert len(opt.history_) == len(opt.err_history_)
    np.testing.assert_allclose(result, expected, atol=1e-2)
    np.testing.assert_array_equal(result == 0, expected == 0)


@pytest.mark.parametrize(
    ["errs", "expected"],
    (([3, 1, 0.9], 1), ([1, 0, 0], 1)),
    ids=["basic", "zero-error"],
)
def test_ssr_inflection(errs, expected):
    result = _ind_inflection(errs)
    assert result == expected


@pytest.mark.parametrize(
    ["errs", "expected", "message"],
    (
        ([1], ValueError, "single point"),
        ([-1, 1, 1], ValueError, ""),
    ),
    ids=["length-1", "negative"],
)
def test_ssr_inflection_bad_args(errs, expected, message):
    with pytest.raises(expected, match=message):
        _ind_inflection(errs)



from scipy.signal import savgol_filter

from .finite_difference import FiniteDifference


class SmoothedFiniteDifference(FiniteDifference):
    """Smoothed finite difference derivatives.

    Perform differentiation by smoothing input data then applying a finite
    difference method.

    Parameters
    ----------
    smoother: function, optional (default :code:`savgol_filter`)
        Function to perform smoothing. Must be compatible with the
        following call signature: :code:`x_smoothed = smoother(x, **smoother_kws)`

    smoother_kws: dict, optional (default :code:`{}`)
        Arguments passed to smoother when it is invoked.

    save_smooth: bool
        Whether to save the smoothed coordinate values or not.

    **kwargs: kwargs
        Additional parameters passed to the :meth:`pysindy.FiniteDifference.__init__`
        function.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.differentiation import SmoothedFiniteDifference
    >>> t = np.linspace(0,1,10)
    >>> X = np.vstack((np.sin(t),np.cos(t))).T
    >>> sfd = SmoothedFiniteDifference(smoother_kws={'window_length': 5})
    >>> sfd._differentiate(X, t)
    array([[ 1.00013114e+00,  7.38006789e-04],
           [ 9.91779070e-01, -1.10702304e-01],
           [ 9.73376491e-01, -2.20038119e-01],
           [ 9.43001496e-01, -3.26517615e-01],
           [ 9.00981354e-01, -4.29066632e-01],
           [ 8.47849424e-01, -5.26323977e-01],
           [ 7.84260982e-01, -6.17090177e-01],
           [ 7.11073255e-01, -7.00180971e-01],
           [ 6.29013295e-01, -7.74740601e-01],
           [ 5.39752150e-01, -8.41980082e-01]])
    """

    def __init__(
        self, smoother=savgol_filter, smoother_kws={}, save_smooth=True, **kwargs
    ):
        super(SmoothedFiniteDifference, self).__init__(**kwargs)
        self.smoother = smoother
        self.smoother_kws = smoother_kws
        self.save_smooth = save_smooth

        if smoother is savgol_filter:
            if "window_length" not in smoother_kws:
                self.smoother_kws["window_length"] = 11
            if "polyorder" not in smoother_kws:
                self.smoother_kws["polyorder"] = 3
            self.smoother_kws["axis"] = 0

    def _differentiate(self, x, t):
        """Apply finite difference method after smoothing."""
        x_smooth = self.smoother(x, **self.smoother_kws)
        x_dot = super(SmoothedFiniteDifference, self)._differentiate(x_smooth, t)
        if self.save_smooth:
            self.smoothed_x_ = x_smooth
        else:
            self.smoothed_x_ = x
        return x_dot



from itertools import repeat
from typing import Optional
from typing import Sequence
from warnings import warn

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ..utils import AxesArray
from .base import _unique
from .base import BaseFeatureLibrary
from .base import x_sequence_or_item
from .weak_pde_library import WeakPDELibrary


class GeneralizedLibrary(BaseFeatureLibrary):
    """Put multiple libraries into one library. All settings
    provided to individual libraries will be applied. Note that this class
    allows one to specifically choose which input variables are used for
    each library, and take tensor products of any pair of libraries. Tensored
    libraries inherit the same input variables specified for the individual
    libraries.

    Parameters
    ----------
    libraries : list of libraries
        Library instances to be applied to the input matrix.

    tensor_array : 2D list of booleans, optional, (default None)
        Default is to not tensor any of the libraries together. Shape
        equal to the # of tensor libraries and the # feature libraries.
        Indicates which pairs of libraries to tensor product together and
        add to the overall library. For instance if you have 5 libraries,
        and want to do two tensor products, you could use the list
        [[1, 0, 0, 1, 0], [0, 1, 0, 1, 1]] to indicate that you want two
        tensored libraries from tensoring libraries 0 and 3 and libraries
        1, 3, and 4.

    inputs_per_library : Sequence of Seqeunces of ints (default None)
        list that specifies which input indexes should be passed as
        inputs for each of the individual feature libraries.
        length must equal the number of feature libraries.  Default is
        that all inputs are used for every library.

    Attributes
    ----------
    self.libraries_full_: list[BaseFeatureLibrary]
        The fitted libraries

    n_features_in_ : int
        The total number of input features.

    n_output_features_ : int
        The total number of output features. The number of output features
        is the sum of the numbers of output features for each of the
        concatenated libraries.

    Examples
    --------
    >>> import numpy as np
    >>> from pysindy.feature_library import FourierLibrary, CustomLibrary
    >>> from pysindy.feature_library import GeneralizedLibrary
    >>> x = np.array([[0.,-1],[1.,0.],[2.,-1.]])
    >>> functions = [lambda x : np.exp(x), lambda x,y : np.sin(x+y)]
    >>> lib_custom = CustomLibrary(library_functions=functions)
    >>> lib_fourier = FourierLibrary()
    >>> lib_generalized = GeneralizedLibrary([lib_custom, lib_fourier])
    >>> lib_generalized.fit(x)
    >>> lib_generalized.transform(x)
    """

    def __init__(
        self,
        libraries: list,
        tensor_array=None,
        inputs_per_library: Optional[Sequence[Sequence[int]]] = None,
        exclude_libraries=[],
    ):
        if len(libraries) > 0:
            self.libraries = libraries

            if has_weak(self) and has_nonweak(self):
                raise ValueError(
                    "At least one of the libraries is a weak form library, "
                    "and at least one of the libraries is not, which will "
                    "result in a nonsensical optimization problem. Please use "
                    "all weak form libraries or no weak form libraries."
                )
        else:
            raise ValueError(
                "Empty or nonsensical library list passed to this library."
            )
        if inputs_per_library is not None:
            if len(inputs_per_library) != len(libraries):
                raise ValueError(
                    "If specifying different inputs for each library, then "
                    "first dimension of inputs_per_library must be equal to "
                    "the number of libraries being used."
                )
            if isinstance(inputs_per_library, np.ndarray):
                warn(
                    "inputs_per_library should no longer be passed as a numpy array",
                    UserWarning,
                )
                inputs_per_library = [list(row) for row in inputs_per_library]
            if any(x_ind < 0 for inputs in inputs_per_library for x_ind in inputs):
                raise ValueError(
                    "The inputs_per_library parameter must be a numpy array "
                    "of integers with values between 0 and "
                    "len(input_variables) - 1."
                )

        if tensor_array is not None:
            if np.asarray(tensor_array).ndim != 2:
                raise ValueError("Tensor product array should be 2D list.")
            if np.asarray(tensor_array).shape[-1] != len(libraries):
                raise ValueError(
                    "If specifying tensor products between libraries, then "
                    "last dimension of tensor_array must be equal to the "
                    "number of libraries being used."
                )
            if np.any(np.ravel(tensor_array) > 1) or np.any(np.ravel(tensor_array) < 0):
                raise ValueError(
                    "The tensor_array parameter must be a numpy array "
                    "of booleans, so values must be either 0 or 1."
                )
            for i in range(len(tensor_array)):
                if np.sum(tensor_array[i]) < 2:
                    raise ValueError(
                        "If specifying libraries to tensor together, must "
                        "specify at least two libraries (there should be at "
                        "least two entries with value 1 in the tensor_array)."
                    )
        self.tensor_array = tensor_array
        self.inputs_per_library = inputs_per_library
        self.exclude_libraries = exclude_libraries

    @x_sequence_or_item
    def fit(self, x_full, y=None):
        """
        Compute number of output features.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            The data.

        Returns
        -------
        self : instance
        """
        n_features = x_full[0].shape[x_full[0].ax_coord]
        self.n_features_in_ = n_features

        # If parameter is not set, use all the inputs
        if self.inputs_per_library is None:
            self.inputs_per_library = list(
                repeat(list(range(n_features)), len(self.libraries))
            )
        else:
            # Check that the numbers in inputs_per_library are sensible
            if any(
                input_ind >= n_features
                for input_list in self.inputs_per_library
                for input_ind in input_list
            ):
                raise ValueError(
                    "Each row in inputs_per_library must consist of integers "
                    "between 0 and the number of total input features - 1. "
                )

        # First fit all libraries separately below, with subset of the inputs
        fitted_libs = [
            lib.fit([x[..., _unique(self.inputs_per_library[i])] for x in x_full], y)
            for i, lib in enumerate(self.libraries)
        ]

        # Next, tensor some libraries and append them to the list
        if self.tensor_array is not None:
            num_tensor_prods = np.shape(self.tensor_array)[0]
            for i in range(num_tensor_prods):
                lib_inds = np.ravel(np.where(self.tensor_array[i]))
                library_subset = np.asarray(fitted_libs)[lib_inds]
                library_full = np.prod(library_subset)
                library_full._set_inputs_per_library(
                    [self.inputs_per_library[lib_ind] for lib_ind in lib_inds]
                )
                library_full.fit(x_full, y)
                fitted_libs.append(library_full)

        # Calculate the sum of output features
        self.n_output_features_ = sum(
            lib.n_output_features_
            for lib in fitted_libs
            if lib not in self.exclude_libraries
        )

        # Save fitted libs
        self.libraries_full_ = fitted_libs

        return self

    @x_sequence_or_item
    def transform(self, x_full):
        """Transform data with libs provided below.

        Parameters
        ----------
        x : array-like, shape [n_samples, n_features]
            The data to transform, row by row.

        Returns
        -------
        xp : np.ndarray, shape [n_samples, NP]
            The matrix of features, where NP is the number of features
            generated from applying the custom functions to the inputs.

        """
        check_is_fitted(self, attributes=["n_features_in_"])

        xp_full = []
        for x in x_full:
            n_features = x.shape[x.ax_coord]
            n_input_features = self.n_features_in_
            if n_features != n_input_features:
                raise ValueError("x shape does not match training shape")

            xps = []
            for i, lib in enumerate(self.libraries_full_):
                if i < len(self.inputs_per_library):
                    if i not in self.exclude_libraries:
                        xps.append(
                            lib.transform(
                                [x[..., _unique(self.inputs_per_library[i])]]
                            )[0]
                        )
                else:
                    xps.append(lib.transform([x])[0])

            xp = AxesArray(np.concatenate(xps, axis=xps[0].ax_coord), xps[0].axes)
            xp_full = xp_full + [xp]
        return xp_full

    def get_feature_names(self, input_features=None):
        """Return feature names for output features.

        Parameters
        ----------
        input_features : list of string, length n_features, optional
            String names for input features if available. By default,
            "x0", "x1", ... "xn_features" is used.

        Returns
        -------
        output_feature_names : list of string, length n_output_features
        """
        check_is_fitted(self)
        feature_names = list()
        for i, lib in enumerate(self.libraries_full_):
            if i not in self.exclude_libraries:
                if i < len(self.libraries):
                    if input_features is None:
                        input_features_i = [
                            "x%d" % k for k in _unique(self.inputs_per_library[i])
                        ]
                    else:
                        input_features_i = np.asarray(input_features)[
                            _unique(self.inputs_per_library[i])
                        ].tolist()
                else:
                    # Tensor libraries need all the inputs and then internally
                    # handle the subsampling of the input variables
                    if input_features is None:
                        input_features_i = ["x{k}" for k in range(self.n_features_in_)]
                    else:
                        input_features_i = input_features
                feature_names += lib.get_feature_names(input_features_i)
        return feature_names

    def calc_trajectory(self, diff_method, x, t):
        return self.libraries[0].calc_trajectory(diff_method, x, t)

    def get_spatial_grid(self):
        for lib_k in self.libraries:
            spatial_grid = lib_k.get_spatial_grid()
            if spatial_grid is not None:
                return spatial_grid


def has_weak(lib):
    if isinstance(lib, WeakPDELibrary):
        return True
    elif hasattr(lib, "libraries_"):
        for lib_k in lib.libraries_:
            if has_weak(lib_k):
                return True
    return False


def has_nonweak(lib):
    if hasattr(lib, "libraries_"):
        for lib_k in lib.libraries_:
            if has_nonweak(lib_k):
                return True
    elif not isinstance(lib, WeakPDELibrary):
        return True
    return False



from __future__ import annotations

import warnings
from typing import Tuple

try:
    import cvxpy as cp

    cvxpy_flag = True
except ImportError:
    cvxpy_flag = False
    pass
import numpy as np
from scipy.linalg import cho_factor
from sklearn.exceptions import ConvergenceWarning

from ..utils import reorder_constraints
from ._constrained_sr3 import ConstrainedSR3


class StableLinearSR3(ConstrainedSR3):
    """
    Sparse relaxed regularized regression for building a-priori
    stable linear models. This requires making a matrix negative definite,
    which can be challenging. Here we use a similar method to the
    TrappingOptimizer algorithm. Linear equality and linear inequality
    constraints are both allowed, as in the ConstrainedSR3 optimizer.

    Attempts to minimize the objective function

    .. math::

        0.5\\|y-Xw\\|^2_2 + \\lambda R(u)
        + (0.5 / \\nu)\\|w-u\\|^2_2

    .. math::

        \\text{subject to } Cu = d, Du = e, w negative definite

    over u and w, where :math:`R(u)` is a regularization function, C and D are
    constraint matrices, and d and e are vectors of values.
    NOTE: This optimizer is intended for building purely linear models that
    are guaranteed to be stable.

    Parameters
    ----------
    regularizer : string, optional (default 'l1')
        Regularization function to use. Currently implemented options
        are 'l1' (l1 norm), 'l2' (l2 norm), 'weighted_l1' (weighted l1 norm),
        and 'weighted_l2' (weighted l2 norm).
        Note that the regularizer must be convex here.
    constraint_lhs : numpy ndarray, optional (default None)
        Shape should be (n_constraints, n_features * n_targets),
        The left hand side matrix C of Cw <= d.
        There should be one row per constraint.

    constraint_rhs : numpy ndarray, shape (n_constraints,), optional (default None)
        The right hand side vector d of Cw <= d.

    constraint_order : string, optional (default "target")
        The format in which the constraints ``constraint_lhs`` were passed.
        Must be one of "target" or "feature".
        "target" indicates that the constraints are grouped by target:
        i.e. the first ``n_features`` columns
        correspond to constraint coefficients on the library features
        for the first target (variable), the next ``n_features`` columns to
        the library features for the second target (variable), and so on.
        "feature" indicates that the constraints are grouped by library
        feature: the first ``n_targets`` columns correspond to the first
        library feature, the next ``n_targets`` columns to the second library
        feature, and so on.

    inequality_constraints : bool, optional (default False)
        If True, CVXPY methods are used to solve the problem.

    verbose_cvxpy : bool, optional (default False)
        Boolean flag which is passed to CVXPY solve function to indicate if
        output should be verbose or not. Only relevant for optimizers that
        use the CVXPY package in some capabity.

    See base class for additional arguments

    Attributes
    ----------
    coef_ : array, shape (n_features,) or (n_targets, n_features)
        Regularized weight vector(s). This is the v in the objective
        function.

    coef_full_ : array, shape (n_features,) or (n_targets, n_features)
        Weight vector(s) that are not subjected to the regularization.
        This is the w in the objective function.
    """

    def __init__(
        self,
        reg_weight_lam=0.1,
        regularizer="l1",
        relax_coeff_nu=1.0,
        tol=1e-5,
        max_iter=30,
        trimming_fraction=0.0,
        trimming_step_size=1.0,
        constraint_lhs=None,
        constraint_rhs=None,
        constraint_order="target",
        normalize_columns=False,
        copy_X=True,
        initial_guess=None,
        equality_constraints=False,
        inequality_constraints=False,
        constraint_separation_index=0,
        verbose=False,
        verbose_cvxpy=False,
        gamma=-1e-8,
        unbias=False,
    ):
        super().__init__(
            reg_weight_lam=reg_weight_lam,
            regularizer=regularizer,
            relax_coeff_nu=relax_coeff_nu,
            tol=tol,
            trimming_fraction=trimming_fraction,
            trimming_step_size=trimming_step_size,
            max_iter=max_iter,
            initial_guess=initial_guess,
            copy_X=copy_X,
            normalize_columns=normalize_columns,
            verbose=verbose,
            verbose_cvxpy=verbose_cvxpy,
            constraint_lhs=constraint_lhs,
            constraint_rhs=constraint_rhs,
            constraint_order=constraint_order,
            equality_constraints=equality_constraints,
            inequality_constraints=inequality_constraints,
            constraint_separation_index=constraint_separation_index,
            unbias=unbias,
        )
        self.gamma = gamma
        self.alpha_A = relax_coeff_nu
        self.max_iter = max_iter
        warnings.warn(
            "This optimizer is set up to only be used with a purely linear"
            " library in the variables. No constant or nonlinear terms!"
        )
        if not np.isclose(reg_weight_lam, 0.0):
            warnings.warn(
                "This optimizer uses CVXPY if the reg_weight_lam is nonzero, "
                " meaning the optimization will be much slower for large "
                "datasets."
            )

    def _create_var_and_part_cost(
        self,
        x: np.ndarray,
        y: np.ndarray,
        coef_sparse: np.ndarray,
        coef_neg_def: np.ndarray,
    ) -> Tuple["cp.Variable", "cp.Expression"]:
        xi = cp.Variable(coef_sparse.shape[0] * coef_sparse.shape[1])
        cost = cp.sum_squares(x @ xi - y.flatten())
        cost = cost + cp.sum_squares(xi - coef_neg_def.flatten()) / (
            2 * self.relax_coeff_nu
        )
        penalty = self._calculate_penalty(
            self.regularizer, np.ravel(self.reg_weight_lam), xi
        )
        return xi, cost + penalty

    def _update_coef_cvxpy(self, x, y, coef_sparse, coef_negative_definite):
        """
        Update the coefficients using CVXPY. This function is called if
        the sparsity weight is nonzero or constraints are used.
        """
        xi, cost = self._create_var_and_part_cost(
            x, y, coef_sparse, coef_negative_definite
        )
        if self.use_constraints:
            if self.inequality_constraints and self.equality_constraints:
                # Process equality constraints then inequality constraints
                prob = cp.Problem(
                    cp.Minimize(cost),
                    [
                        self.constraint_lhs[: self.constraint_separation_index, :] @ xi
                        <= self.constraint_rhs[: self.constraint_separation_index],
                        self.constraint_lhs[self.constraint_separation_index :, :] @ xi
                        == self.constraint_rhs[self.constraint_separation_index :],
                    ],
                )
            elif self.inequality_constraints:
                prob = cp.Problem(
                    cp.Minimize(cost),
                    [self.constraint_lhs @ xi <= self.constraint_rhs],
                )
            else:
                prob = cp.Problem(
                    cp.Minimize(cost),
                    [self.constraint_lhs @ xi == self.constraint_rhs],
                )
        else:
            prob = cp.Problem(cp.Minimize(cost))

        try:
            prob.solve(
                max_iter=self.max_iter**2,
                eps_abs=self.tol,
                eps_rel=self.tol,
                verbose=self.verbose_cvxpy,
            )
        except cp.error.SolverError:
            print("Solver failed, setting coefs to zeros")
            xi.value = np.zeros(coef_sparse.shape[0] * coef_sparse.shape[1])

        if xi.value is None:
            warnings.warn(
                "Infeasible solve, probably an issue with the regularizer "
                " or the constraint that was used.",
                ConvergenceWarning,
            )
            return None
        coef_new = (xi.value).reshape(coef_sparse.shape)
        return coef_new

    def _update_A(self, A_old, coef_sparse):
        """
        Update the auxiliary variable that approximates the coefficients
        (which is a matrix of linear coefficients). Taken and slightly altered
        from the TrappingOptimizer code.
        """
        r = A_old.shape[1]
        if A_old.shape[0] == r:
            eigvals, eigvecs = np.linalg.eig(A_old.T)
            eigPW, eigvecsPW = np.linalg.eig(coef_sparse.T)
        else:
            eigvals, eigvecs = np.linalg.eig(A_old[:r, :r].T)
            eigPW, eigvecsPW = np.linalg.eig(coef_sparse[:r, :r].T)
        A = np.diag(eigvals)
        for i in range(r):
            if np.real(eigvals[i]) > self.gamma:
                A[i, i] = self.gamma + np.imag(eigvals[i]) * 1j
        if A_old.shape[0] == r:
            return np.real(eigvecsPW @ A @ np.linalg.inv(eigvecsPW))
        else:
            A_temp = np.zeros(A_old.shape)
            A_temp[:r, :r] = np.real(eigvecsPW @ A @ np.linalg.inv(eigvecsPW))
            A_temp[r:, :r] = A_old[r:, :r]
            return A_temp.T

    def _reduce(self, x, y):
        """
        Perform at most ``self.max_iter`` iterations of the SR3 algorithm
        with inequality constraints.

        Assumes initial guess for coefficients is stored in ``self.coef_``.
        """
        if self.initial_guess is not None:
            self.coef_ = self.initial_guess

        coef_sparse = self.coef_.T
        coef_negative_definite = coef_sparse.copy()
        n_samples, n_features = x.shape
        n_targets = y.shape[1]

        if self.use_trimming:
            trimming_array = np.repeat(1.0 - self.trimming_fraction, n_samples)
            self.history_trimming_ = [trimming_array]

        if self.use_constraints and self.constraint_order.lower() == "target":
            self.constraint_lhs = reorder_constraints(self.constraint_lhs, n_features)

        # Precompute some objects for optimization
        H = np.dot(x.T, x) + np.diag(np.full(x.shape[1], 1.0 / self.relax_coeff_nu))
        x_transpose_y = np.dot(x.T, y)
        if not self.use_constraints:
            cho = cho_factor(H)
        x_expanded = np.zeros((n_samples, n_targets, n_features, n_targets))
        for i in range(n_targets):
            x_expanded[:, i, :, i] = x
        x_expanded = np.reshape(
            x_expanded, (n_samples * n_targets, n_targets * n_features)
        )

        # Print initial values for each term in the optimization
        if self.verbose:
            row = [
                "Iteration",
                "|y - Xw|^2",
                "|w-u|^2/v",
                "R(u)",
                "Total Error: |y - Xw|^2 + |w - u|^2 / v + R(u)",
            ]
            print(
                "{: >10} ... {: >10} ... {: >10} ... {: >10} ... {: >10}".format(*row)
            )

        objective_history = []
        eigs_history = []
        coef_history = []
        for k in range(self.max_iter):
            if not np.isclose(self.reg_weight_lam, 0.0) or self.use_constraints:
                coef_sparse = self._update_coef_cvxpy(
                    x_expanded, y, coef_sparse, coef_negative_definite
                )
            else:
                coef_sparse = self._update_full_coef(
                    cho, x_transpose_y, coef_negative_definite
                )
            coef_negative_definite = self._update_A(
                coef_negative_definite
                - self.alpha_A
                * (coef_negative_definite - coef_sparse)
                / self.relax_coeff_nu,
                coef_sparse,
            ).T
            objective_history.append(
                self._objective(x, y, k, coef_negative_definite, coef_sparse)
            )
            eigs_history.append(np.sort(np.linalg.svd(coef_sparse, compute_uv=False)))
            coef_history.append(coef_sparse)
            if self._convergence_criterion() < self.tol:
                # TODO: Update this for trimming/constraints
                break
        else:
            warnings.warn(
                f"StableLinearSR3 did not converge after {self.max_iter} iterations.",
                ConvergenceWarning,
            )

        if self.use_constraints and self.constraint_order.lower() == "target":
            self.constraint_lhs = reorder_constraints(
                self.constraint_lhs, n_features, output_order="target"
            )
        self.coef_ = coef_sparse.T
        self.coef_full_ = coef_negative_definite.T
        if self.use_trimming:
            self.trimming_array = trimming_array
        self.objective_history = objective_history
        self.eigs_history = np.array(eigs_history)
        self.coef_history = np.array(coef_history)



"""
Shared pytest fixtures for unit tests.
"""
from pathlib import Path
from tempfile import gettempdir

import jax
import numpy as np
import pytest
from scipy.integrate import solve_ivp

from pysindy.differentiation import FiniteDifference
from pysindy.differentiation import SpectralDerivative
from pysindy.feature_library import CustomLibrary
from pysindy.feature_library import FourierLibrary
from pysindy.feature_library import GeneralizedLibrary
from pysindy.feature_library import PDELibrary
from pysindy.feature_library import PolynomialLibrary
from pysindy.utils.odes import logistic_map
from pysindy.utils.odes import logistic_map_control
from pysindy.utils.odes import logistic_map_multicontrol
from pysindy.utils.odes import lorenz
from pysindy.utils.odes import lorenz_control

jax.config.update("jax_platform_name", "cpu")


def pytest_addoption(parser):
    parser.addoption(
        "--external-notebook",
        action="append",
        default=[],
        help=(
            "name of notebook to test.  Only valid if running"
            " test_notebooks.test_external"
        ),
    )
    parser.addoption(
        "--jax-pcc",
        action="store_true",
        help=(
            "Whether to cache @jax.jit compilations to disk."
            "It speeds up the slowest tests appx 2x, writing files to /tmp/jax_cache"
        ),
    )


def pytest_generate_tests(metafunc):
    if "external_notebook" in metafunc.fixturenames:
        metafunc.parametrize(
            "external_notebook",
            [
                Path(f.lstrip('"').rstrip('"'))
                for f in metafunc.config.getoption("external_notebook")
            ],
        )


@pytest.fixture(scope="session")
def set_jax_pcc_env(request):
    if request.config.getoption("--jax-pcc"):
        cache_dir = Path(gettempdir()) / "jax_cache/"
        cache_dir.mkdir(exist_ok=True)
        jax.config.update("jax_compilation_cache_dir", str(cache_dir))
        jax.config.update("jax_persistent_cache_min_entry_size_bytes", -1)
        jax.config.update("jax_persistent_cache_min_compile_time_secs", -1)
        jax.config.update("jax_persistent_cache_enable_xla_caches", "all")


@pytest.fixture(scope="session")
def data_1d():
    t = np.linspace(0, 1, 10)
    x = 2 * t.reshape(-1, 1)
    return x, t


@pytest.fixture(scope="session")
def data_1d_bad_shape():
    t = np.linspace(0, 5, 10)
    x = 2 * t
    return x, t


@pytest.fixture(scope="session")
def data_lorenz():
    t = np.linspace(0, 1, 12)
    x0 = [8, 27, -7]
    x = solve_ivp(lorenz, (t[0], t[-1]), x0, t_eval=t).y.T

    return x, t


@pytest.fixture
def data_multiple_trajectories():
    n_points = [100, 200, 500]
    initial_conditions = [
        [8, 27, -7],
        [-10.9595724, 21.7346758, 24.5722540],
        [-3.95406365, -9.21825124, 12.07459147],
    ]

    x_list = []
    t_list = []
    for n, x0 in zip(n_points, initial_conditions):
        t = np.linspace(0, 5, n)
        t_list.append(t)
        x_list.append(solve_ivp(lorenz, (t[0], t[-1]), x0, t_eval=t).y.T)

    return x_list, t_list


@pytest.fixture(scope="session")
def diffuse_multiple_trajectories():
    def diffuse(t, u, dx, nx):
        u = np.reshape(u, nx)
        du = SpectralDerivative(d=2, axis=0)._differentiate(u, dx)
        return np.reshape(u * du, nx)

    # Required for accurate solve_ivp results
    integrator_keywords = {}
    integrator_keywords["rtol"] = 1e-8
    integrator_keywords["method"] = "LSODA"
    integrator_keywords["atol"] = 1e-8
    N = 25
    h0 = 1.0
    L = 5
    T = 1
    t = np.linspace(0, T, N)
    x = np.arange(0, N) * L / N
    ep = 0.5 * h0
    y0 = np.reshape(
        h0 + ep * np.cos(4 * np.pi / L * x) + ep * np.cos(2 * np.pi / L * x), N
    )
    dx = x[1] - x[0]
    sol1 = solve_ivp(
        diffuse, (t[0], t[-1]), y0=y0, t_eval=t, args=(dx, N), **integrator_keywords
    )
    u = [np.reshape(sol1.y, (N, N, 1))]
    return t, x, u


@pytest.fixture(scope="session")
def data_discrete_time():
    n_steps = 100
    mu = 3.6
    x = np.zeros((n_steps))
    x[0] = 0.5
    for i in range(1, n_steps):
        x[i] = logistic_map(x[i - 1], mu)

    return x


@pytest.fixture(scope="session")
def data_discrete_time_multiple_trajectories():
    n_steps = 100
    mus = [1, 2.3, 3.6]
    x = [np.zeros((n_steps)) for mu in mus]
    for i, mu in enumerate(mus):
        x[i][0] = 0.5
        for k in range(1, n_steps):
            x[i][k] = logistic_map(x[i][k - 1], mu)

    return x


@pytest.fixture(scope="session")
def data_1d_random_pde():
    n = 100
    t = np.linspace(0, 10, n)
    dt = t[1] - t[0]
    x = np.linspace(0, 10, n)
    u = np.random.randn(n, n, 1)
    u_dot = FiniteDifference(axis=1)._differentiate(u, t=dt)
    return t, x, u, u_dot


@pytest.fixture(scope="session")
def data_2d_random_pde():
    n = 4
    t = np.linspace(0, 10, n)
    dt = t[1] - t[0]
    x = np.linspace(0, 10, n)
    y = np.linspace(0, 10, n)
    X, Y = np.meshgrid(x, y)
    spatial_grid = np.asarray([X, Y]).T
    u = np.random.randn(n, n, n, 2)
    u_dot = FiniteDifference(axis=2)._differentiate(u, t=dt)
    return spatial_grid, u, u_dot


@pytest.fixture(scope="session")
def data_3d_random_pde():
    n = 4
    t = np.linspace(0, 10, n)
    dt = t[1] - t[0]
    x = np.linspace(0, 10, n)
    y = np.linspace(0, 10, n)
    z = np.linspace(0, 10, n)
    (
        X,
        Y,
        Z,
    ) = np.meshgrid(x, y, z, indexing="ij")
    spatial_grid = np.asarray([X, Y, Z])
    spatial_grid = np.transpose(spatial_grid, axes=[1, 2, 3, 0])
    u = np.random.randn(n, n, n, n, 2)
    u_dot = FiniteDifference(axis=3)._differentiate(u, t=dt)
    return spatial_grid, u, u_dot


@pytest.fixture(scope="session")
def data_5d_random_pde():
    n = 4
    t = np.linspace(0, n, n)
    dt = t[1] - t[0]
    v = np.linspace(0, 10, n)
    w = np.linspace(0, 10, n)
    x = np.linspace(0, 10, n)
    y = np.linspace(0, 10, n)
    z = np.linspace(0, 10, n)
    V, W, X, Y, Z = np.meshgrid(v, w, x, y, z, indexing="ij")
    spatial_grid = np.asarray([V, W, X, Y, Z])
    spatial_grid = np.transpose(spatial_grid, axes=[1, 2, 3, 4, 5, 0])
    u = np.random.randn(n, n, n, n, n, n, 2)
    u_dot = FiniteDifference(axis=5)._differentiate(u, t=dt)
    return spatial_grid, u, u_dot


@pytest.fixture(scope="session")
def data_2d_resolved_pde():
    n = 8
    nt = 1000
    t = np.linspace(0, 10, nt)
    dt = t[1] - t[0]
    x = np.linspace(0, 10, n)
    y = np.linspace(0, 10, n)
    X, Y = np.meshgrid(x, y)
    spatial_grid = np.asarray([X, Y]).T
    u = np.random.randn(n, n, nt, 2)
    u_dot = FiniteDifference(axis=-2)._differentiate(u, t=dt)
    return spatial_grid, u, u_dot


@pytest.fixture(scope="session")
def data_derivative_1d():
    x = 2 * np.linspace(1, 100, 100)
    x_dot = 2 * np.ones(100)
    return x, x_dot


@pytest.fixture(scope="session")
def data_derivative_quasiperiodic_1d():
    t = np.arange(1000) * 2 * np.pi / 1000
    x = 2 * np.sin(t)
    x_dot = 2 * np.cos(t)
    return t, x, x_dot


@pytest.fixture(scope="session")
def data_derivative_2d():
    x = np.zeros((100, 2))
    x[:, 0] = 2 * np.linspace(1, 100, 100)
    x[:, 1] = -10 * np.linspace(1, 100, 100)

    x_dot = np.ones((100, 2))
    x_dot[:, 0] *= 2
    x_dot[:, 1] *= -10
    return x, x_dot


@pytest.fixture(scope="session")
def data_derivative_quasiperiodic_2d():
    t = np.arange(1000) * 2 * np.pi / 1000
    x = np.zeros((1000, 2))
    x[:, 0] = 2 * np.sin(t)
    x[:, 1] = 2 * np.cos(2 * t)
    x_dot = np.zeros((1000, 2))
    x_dot[:, 0] = 2 * np.cos(t)
    x_dot[:, 1] = -4 * np.sin(2 * t)
    return t, x, x_dot


@pytest.fixture(scope="session")
def data_2dspatial():
    u = np.zeros((100, 50, 2))
    x = np.linspace(1, 100, 100)
    y = np.linspace(1, 50, 50)
    X, Y = np.meshgrid(x, y, indexing="ij")
    u[:, :, 0] = np.cos(X) * np.sin(Y)
    u[:, :, 1] = -np.sin(X) * np.cos(Y) ** 2
    return x, y, u


@pytest.fixture
def custom_library():
    library_functions = [
        lambda x: x,
        lambda x: x**2,
        lambda x: 0 * x,
        lambda x, y: x * y,
    ]
    function_names = [
        lambda s: str(s),
        lambda s: str(s) + "^2",
        lambda s: "0",
        lambda s, t: str(s) + " " + str(t),
    ]

    return CustomLibrary(
        library_functions=library_functions, function_names=function_names
    )


@pytest.fixture
def custom_library_bias():
    library_functions = [
        lambda x: x,
        lambda x: x**2,
        lambda x: 0 * x,
        lambda x, y: x * y,
    ]
    function_names = [
        lambda s: str(s),
        lambda s: str(s) + "^2",
        lambda s: "0",
        lambda s, t: str(s) + " " + str(t),
    ]

    return CustomLibrary(
        library_functions=library_functions,
        function_names=function_names,
        include_bias=True,
    )


@pytest.fixture
def generalized_library():
    tensor_array = [[1, 1]]
    return GeneralizedLibrary(
        [PolynomialLibrary(), FourierLibrary()],
        tensor_array=tensor_array,
    )


@pytest.fixture
def sindypi_library(data_lorenz):
    function_library = PolynomialLibrary(degree=2, include_bias=False)
    _, t = data_lorenz

    return PDELibrary(
        function_library=function_library,
        temporal_grid=t,
        implicit_terms=True,
        derivative_order=1,
    )


@pytest.fixture
def ode_library():
    function_library = PolynomialLibrary(degree=2, include_bias=False)

    return PDELibrary(
        function_library=function_library,
    )


@pytest.fixture
def pde_library(data_lorenz):
    _, spatial_grid = data_lorenz
    function_library = PolynomialLibrary(degree=2, include_bias=False)

    return PDELibrary(
        function_library=function_library,
        spatial_grid=spatial_grid,
        derivative_order=4,
    )


@pytest.fixture(scope="session")
def data_linear_oscillator_corrupted():
    t = np.linspace(0, 1, 100)
    x = 3 * np.exp(-2 * t)
    y = 0.5 * np.exp(t)
    np.random.seed(1)
    corrupt_idxs = np.random.choice(np.arange(1, t.size - 1), t.size // 20)
    x[corrupt_idxs] = 0
    X = np.stack((x, y), axis=-1)
    X_dot = FiniteDifference(order=2)(X, t)

    # build an array of the indices of samples that should be trimmed
    trimmed_idxs = np.concatenate((corrupt_idxs - 1, corrupt_idxs, corrupt_idxs + 1))
    trimming_array = np.ones(X.shape[0])
    trimming_array[trimmed_idxs] = 0.0

    return X, X_dot, trimming_array


@pytest.fixture(scope="session")
def data_linear_combination():
    t = np.linspace(0, 3, 100).reshape((-1, 1))
    lib = PolynomialLibrary(2, include_bias=False)
    x = lib.fit_transform(np.hstack([t, -2 * t, 3 * t]))
    y = np.stack((x[:, 0] + x[:, 1], x[:, 1] + x[:, 2], x[:, 0] + x[:, 2]), axis=-1)

    return x, y


# Datasets with control inputs


@pytest.fixture(scope="session")
def data_lorenz_c_1d():
    def u_fun(t):
        if len(np.shape(t)) == 0:
            return np.column_stack([np.sin(2 * t), 0])
        else:
            return np.column_stack([np.sin(2 * t), np.zeros(len(t))])

    t = np.linspace(0, 1, 100)
    x0 = [8, 27, -7]
    x = solve_ivp(lorenz_control, (t[0], t[-1]), x0, t_eval=t, args=(u_fun,)).y.T
    u = u_fun(t)

    return x, t, u, u_fun


@pytest.fixture(scope="session")
def data_lorenz_c_2d():
    def u_fun(t):
        return np.column_stack([np.sin(2 * t), t**2])

    t = np.linspace(0, 1, 100)
    x0 = [8, 27, -7]
    x = solve_ivp(lorenz_control, (t[0], t[-1]), x0, t_eval=t, args=(u_fun,)).y.T
    u = u_fun(t)

    return x, t, u, u_fun


@pytest.fixture(scope="session")
def data_discrete_time_c():
    n_steps = 100
    mu = 3.6

    u = 0.01 * np.random.randn(n_steps)
    x = np.zeros((n_steps))
    x[0] = 0.5

    for i in range(1, n_steps):
        x[i] = logistic_map_control(x[i - 1], mu, u[i - 1])

    return x, u


@pytest.fixture(scope="session")
def data_discrete_time_c_multivariable():
    n_steps = 100
    mu = 3.6

    u1 = 0.1 * np.random.randn(n_steps)
    u2 = 0.1 * np.random.randn(n_steps)
    u = np.column_stack((u1, u2))
    x = np.zeros((n_steps))
    x[0] = 0.5
    for i in range(1, n_steps):
        x[i] = logistic_map_multicontrol(x[i - 1], mu, u[i - 1])

    return x, u


@pytest.fixture(scope="session")
def data_discrete_time_multiple_trajectories_c():
    n_steps = 100
    mus = [1, 2.3, 3.6]
    u = [0.001 * np.random.randn(n_steps) for mu in mus]
    x = [np.zeros((n_steps)) for mu in mus]
    for i, mu in enumerate(mus):
        x[i][0] = 0.5
        for k in range(1, n_steps):
            x[i][k] = logistic_map_control(x[i][k - 1], mu, u[i][k - 1])

    return x, u



"""
Wrapper classes for differentiation methods from the :doc:`derivative:index` package.

Some default values used here may differ from those used in :doc:`derivative:index`.
"""
from derivative import methods
from numpy import arange

from .base import BaseDifferentiation


class SINDyDerivative(BaseDifferentiation):
    """
    Wrapper class for differentiation classes from the :doc:`derivative:index` package.
    This class is meant to provide all the same functionality as the
    `dxdt <https://derivative.readthedocs.io/en/latest/api.html\
        #derivative.differentiation.dxdt>`_ method.

    This class also has ``_differentiate`` and ``__call__`` methods which are
    used by PySINDy.

    Parameters
    ----------
    derivative_kws: dictionary, optional
        Keyword arguments to be passed to the
        `dxdt <https://derivative.readthedocs.io/en/latest/api.html\
        #derivative.differentiation.dxdt>`_
        method.

    Notes
    -----
    See the `derivative documentation <https://derivative.readthedocs.io/en/latest/>`_
    for acceptable keywords.
    """

    def __init__(self, axis=0, save_smooth=True, **kwargs):
        self.axis = axis
        self.kwargs = kwargs
        self.save_smooth = save_smooth

    def set_params(self, **params):
        """
        Set the parameters of this estimator.
        Modification of the pysindy method to allow unknown kwargs. This allows using
        the full range of derivative parameters that are not defined as member variables
        in sklearn grid search.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        else:
            self.kwargs.update(params["kwargs"])
            self.save_smooth = params.get("save_smooth", self.save_smooth)

        return self

    def get_params(self, deep=True):
        """Get parameters."""
        params = super().get_params(deep)

        if isinstance(self.kwargs, dict):
            params.update(self.kwargs)
        params["save_smooth"] = self.save_smooth

        return params

    def _differentiate(self, x, t=1):
        if isinstance(t, (int, float)):
            if t < 0:
                raise ValueError("t must be a positive constant or an array")
            t = arange(x.shape[0]) * t

        differentiator = methods[self.kwargs["kind"]](
            **{k: v for k, v in self.kwargs.items() if k != "kind"}
        )
        x_dot = differentiator.d(x, t, axis=self.axis)
        if self.save_smooth:
            self.smoothed_x_ = differentiator.x(x, t, axis=self.axis)
        else:
            self.smoothed_x_ = x
        return x_dot
